from time import perf_counter
import contextlib
import logging
import warnings
from collections import defaultdict
from enum import Enum

from octave_sdk.octave import ClockInfo

from qm import QuantumMachine
from qm.api.models.capabilities import ServerCapabilities
from qm.exceptions import OpenQmException
from qm.octave._calibration_config import (
    _prep_config,
    _get_frequencies,
)
from qm.octave.calibration_db import (
    CalibrationResult,
    octave_output_mixer_name,
)
from octave_sdk import (
    OctaveLOSource,
    RFInputRFSource,
    ClockType,
    ClockFrequency,
    RFOutputMode,
    RFInputLOSource,
    IFMode,
)
from qm.octave.octave_config import _convert_octave_port_to_number, QmOctaveConfig

from typing import Dict, Tuple, Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    from qm.QuantumMachinesManager import QuantumMachinesManager

try:
    from octave_sdk import Octave

    OCTAVE_SDK_LOADED = True

except ModuleNotFoundError:
    # Error handling
    OCTAVE_SDK_LOADED = False
    Octave = None

logger = logging.getLogger(__name__)


class OctaveSDKException(Exception):
    pass


class SetFrequencyException(Exception):
    pass


class ClockMode(Enum):
    Internal = (ClockType.Internal, None)
    External_10MHz = (ClockType.External, ClockFrequency.MHZ_10)
    External_100MHz = (ClockType.External, ClockFrequency.MHZ_100)
    External_1000MHz = (ClockType.Buffered, ClockFrequency.MHZ_1000)


def _run_compiled(compiled_id: str, qm: QuantumMachine):
    from qm.octave._calibration_program import _process_results

    pending_job = qm.queue.add_compiled(compiled_id)
    job = pending_job.wait_for_execution()
    all_done = job.result_handles.wait_for_all_values(timeout=30)
    if not all_done:
        _calibration_failed_error()
    res = _get_results(
        [
            "error",
            "i_track",
            "q_track",
            "c00",
            "c01",
            "c10",
            "c11",
        ],
        job,
    )
    # g_track,
    # phi_track,
    # bool_lo_power_small,
    # bool_lo_step_size_small,
    # bool_too_many_iterations,
    # bool_image_power_small,
    # bool_image_step_size_small,
    # final_g = g_track[-1]
    # final_phi = phi_track[-1]
    # import numpy as np
    # lo = lo
    # if lo[-1] == 0:
    #     lo[-1] = 2 ** -28
    # image = image
    # if image[-1] == 0:
    #     image[-1] = 2 ** -28
    # signal = signal
    # if signal[-1] < 0:
    #     signal[-1] = 8 - signal[-1]
    # exit_lo = (
    #     1 * bool_lo_power_small
    #     + 2 * bool_lo_step_size_small
    #     + 4 * bool_too_many_iterations
    # )
    # print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
    # print(
    #     f"LO suppression is {10 * np.log10(signal[-1] / lo[-1]):.4f} dBc at (I,"
    #     f"Q) = ({final_i:.4f},{final_q:.4f}) "
    # )
    # print(exit_lo)
    # #
    # exit_image = (
    #     1 * bool_image_power_small
    #     + 2 * bool_image_step_size_small
    #     + 4 * bool_too_many_iterations
    # )
    # print(
    #     f"Image suppression is {10 * np.log10(signal[-1] / image[-1]):.4f} dBc "
    #     f"at (g,phi) = ({final_g:.4f},{final_phi:.4f}) "
    # )
    # print(exit_image)
    # print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
    return _process_results(
        res["error"],
        res["i_track"],
        res["q_track"],
        res["c00"],
        res["c01"],
        res["c10"],
        res["c11"],
    )


def _calibration_failed_error(extra: str = None):
    raise Exception(
        f"There was a problem during calibration. please check opx is on"
        f" and valid. If this continues, please contact QM support{extra}"
    )


def _get_results(names, job):
    res = {}
    for name in names:
        if name not in job.result_handles:
            _calibration_failed_error(f"\nerror in result {name}")
        res[name] = job.result_handles.get(name).fetch_all(flat_struct=True)
    return res


def _convert_rf_output_index_to_input_source(rf_output) -> RFInputRFSource:
    if rf_output == 1:
        return RFInputRFSource.Loopback_RF_out_1
    elif rf_output == 2:
        return RFInputRFSource.Loopback_RF_out_2
    elif rf_output == 3:
        return RFInputRFSource.Loopback_RF_out_3
    elif rf_output == 4:
        return RFInputRFSource.Loopback_RF_out_4
    elif rf_output == 5:
        return RFInputRFSource.Loopback_RF_out_5


class OctaveManager:
    def __init__(
        self,
        config: Optional[QmOctaveConfig] = None,
        qmm: Optional["QuantumMachinesManager"] = None,
        capabilities: ServerCapabilities = None,
    ) -> None:

        self._octave_config: QmOctaveConfig
        self._qmm = qmm
        self._capabilities: ServerCapabilities = capabilities

        self._check_and_set_config(config)
        self._initialize()

    def _check_and_set_config(self, config):
        if config and isinstance(config, QmOctaveConfig):
            self._octave_config: QmOctaveConfig = config
        elif config is None:
            self._octave_config: QmOctaveConfig = QmOctaveConfig()
        else:
            raise TypeError("config must be of type QmOctaveConfig")

    def _initialize(self):
        self._octave_clients: Dict[str, Octave] = {}
        devices = self._octave_config.get_devices()

        if devices is not None and len(devices) > 0:
            if not OCTAVE_SDK_LOADED:
                raise OctaveSDKException("Octave sdk is not installed")
            for device_name, connection_info in devices.items():
                loop_backs = self._octave_config.get_lo_loopbacks_by_octave(device_name)
                loop_backs = {input_port: output_port for input_port, output_port in loop_backs.items()}
                self._octave_clients[device_name] = Octave(
                    host=connection_info.host,
                    port=connection_info.port,
                    port_mapping=loop_backs,
                    octave_name=device_name,
                    fan=self._octave_config.fan,
                )

    def set_octave_configuration(self, config: QmOctaveConfig):
        """Args:
        config
        """
        self._check_and_set_config(config)
        self._initialize()

    def get_output_port(
        self,
        opx_i_port: Tuple[str, int],
        opx_q_port: Tuple[str, int],
    ) -> Tuple[str, int]:
        # check both ports are going to the same mixer
        connections = self._octave_config.get_opx_octave_port_mapping()
        i_octave_port = connections[opx_i_port]
        q_octave_port = connections[opx_q_port]
        if i_octave_port[0] != q_octave_port[0]:
            raise Exception("I and Q are not connected to the same octave")
        if i_octave_port[1][-1] != q_octave_port[1][-1]:
            raise Exception("I and Q are not connected to the same octave input")

        return i_octave_port[0], _convert_octave_port_to_number(i_octave_port[1])

    def _get_client(self, octave_port):
        octave = self._octave_clients[octave_port[0]]
        return octave

    def restore_default_state(self, octave_name):
        self._octave_clients[octave_name].restore_default_state()

    def start_batch_mode(self):
        for client in self._octave_clients.values():
            client.start_batch_mode()

    def end_batch_mode(self):
        for client in self._octave_clients.values():
            client.end_batch_mode()

    @contextlib.contextmanager
    def batch_mode(self):

        self.start_batch_mode()
        yield
        self.end_batch_mode()

    def set_clock(
        self,
        octave_name: str,
        clock_type: Optional[ClockType] = None,
        frequency: Optional[ClockFrequency] = None,
        clock_mode: Optional[ClockMode] = None,
    ):

        """This function will set the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        Args:
            octave_name: str
            clock_type: ClockType
            frequency: ClockFrequency
            clock_mode: ClockMode

        Returns:

        """
        if clock_mode is None:
            warnings.warn(
                "set_clock is changing its API, and the 'clock_type' and 'frequency' arguments will be "
                "removed in the next version, please use the 'mode' parameter.",
                category=DeprecationWarning,
            )
        else:
            clock_type, frequency = clock_mode.value

        self._octave_clients[octave_name].set_clock(clock_type, frequency)
        self._octave_clients[octave_name].save_default_state(only_clock=True)

    def get_clock(self, octave_name: str) -> ClockInfo:
        """Return the octave clock type and frequency

        Args:
            octave_name: str

        Returns:
            ClockInfo
        """
        clock = self._octave_clients[octave_name].get_clock()
        return ClockInfo(clock.clock_type, clock.frequency)

    def set_lo_frequency(
        self,
        octave_output_port: Tuple[str, int],
        lo_frequency: float,
        lo_source: OctaveLOSource = OctaveLOSource.Internal,
        set_source: bool = True,
    ):
        """Sets the LO frequency of the synthesizer associated to element

        Args:
            octave_output_port
            lo_frequency
            lo_source
            set_source
        """
        octave = self._get_client(octave_output_port)
        octave_name, port_index = octave_output_port

        loop_backs = self._octave_config.get_lo_loopbacks_by_octave(octave_name)

        if lo_source != OctaveLOSource.Internal and lo_source not in loop_backs:
            raise SetFrequencyException(f"Cannot set frequency to an external lo source" f" {lo_source.name}")

        if set_source:
            octave.rf_outputs[port_index].set_lo_source(lo_source)

        octave.rf_outputs[port_index].set_lo_frequency(lo_source, lo_frequency)
        # octave.save_default_state()

    def set_lo_source(self, octave_output_port: Tuple[str, int], lo_port: OctaveLOSource):
        """Sets the source of LO going to the upconverter associated with element.

        Args:
            octave_output_port
            lo_port
        """
        octave = self._get_client(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_lo_source(lo_port)
        # octave.save_default_state()

    def set_rf_output_mode(self, octave_output_port: Tuple[str, int], switch_mode: RFOutputMode):
        """Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'always_on', 'always_off', 'normal' or 'inverted'
        When in 'normal' mode a high trigger will turn the switch on and a low
        trigger will turn it off
        When in 'inverted' mode a high trigger will turn the switch off and a low
        trigger will turn it on
        When in 'always_on' the switch will be permanently on. When in 'always_off'
        mode the switch will be permanently off.

        Args:
            octave_output_port
            switch_mode
        """
        octave = self._get_client(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_output(switch_mode)
        # octave.save_default_state()

    def set_rf_output_gain(
        self,
        octave_output_port: Tuple[str, int],
        gain_in_db: float,
        lo_frequency: Optional[float] = None,
    ):
        """Sets the RF output gain for the upconverter associated with element.
        if no lo_frequency is given, and lo source is internal, will use the
        internal frequency

        Args:
            octave_output_port
            gain_in_db
            lo_frequency
        """
        octave = self._get_client(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_gain(gain_in_db, lo_frequency)
        # octave.save_default_state()

    def set_downconversion_lo_source(
        self,
        octave_input_port: Tuple[str, int],
        lo_source: RFInputLOSource,
        lo_frequency: Optional[float] = None,
        disable_warning: Optional[bool] = False,
    ):
        """Sets the LO source for the downconverters.

        Args:
            octave_input_port
            lo_source
            lo_frequency
            disable_warning
        """
        octave = self._get_client(octave_input_port)
        octave.rf_inputs[octave_input_port[1]].set_lo_source(lo_source)
        octave.rf_inputs[octave_input_port[1]].set_rf_source(RFInputRFSource.RF_in)
        internal = lo_source == RFInputLOSource.Internal or lo_source == RFInputLOSource.Analyzer
        if lo_frequency is not None and internal:
            octave.rf_inputs[octave_input_port[1]].set_lo_frequency(source_name=lo_source, frequency=lo_frequency)
        # octave.save_default_state()

    def set_downconversion_if_mode(
        self,
        octave_input_port: Tuple[str, int],
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning: Optional[bool] = False,
    ):
        """Sets the IF downconversion stage.
        if_mode can be one of: 'direct', 'mixer', 'envelope_DC', 'envelope_AC','OFF'
        If only one value is given the setting is applied to both IF channels
        (I and Q) for the downconverter associated to element
        (how will we know that? shouldn't this be per downconverter?)
        If if_mode is a tuple, then the IF stage will be assigned to each
        quadrature independently, i.e.:
        if_mode = ('direct', 'envelope_AC') will set the I channel to be
        direct and the Q channel to be 'envelope_AC'

        Args:
            disable_warning
            octave_input_port
            if_mode_q
            if_mode_i
        """

        octave = self._get_client(octave_input_port)
        octave.rf_inputs[octave_input_port[1]].set_if_mode_i(if_mode_i)
        octave.rf_inputs[octave_input_port[1]].set_if_mode_q(if_mode_q)

    def reset(self, octave_name: str) -> bool:
        """
        Will reset the entire Octave HW to default off state
        Warning, will block the code until reset completes

        :param octave_name: str
        :return: True on success, False otherwise
        """
        if self._capabilities.supports_octave_reset:
            return self._octave_clients[octave_name].reset()
        else:
            logger.error("QOP version do not support Octave reset function")
            return False

    def calibrate(
        self,
        octave_output_port: Tuple[str, int],
        lo_if_frequencies_tuple_list: List[Tuple] = None,
        save_to_db=True,
        close_open_quantum_machines=True,
        optimizer_parameters: Optional[dict] = None,
        **kwargs,
    ) -> Dict[Tuple[int, int], CalibrationResult]:
        """calibrates IQ mixer associated with element

        Args:
            close_open_quantum_machines: Boolean, if true (default) all
                running QMs
            octave_output_port
            lo_if_frequencies_tuple_list: A list of LO/IF frequencies
                for which the calibration is to be performed [(LO1,
                IF1), (LO2, IF2), ...]
            save_to_db
            optimizer_parameters
        will be closed for the calibration. Otherwise,
        calibration might fail if there are not enough resources for the calibration
        """
        calibration_start = perf_counter()
        if kwargs:
            logger.warning(f"unused kwargs: {list(kwargs)}, please remove them.")

        octave = self._get_client(octave_output_port)
        output_port = octave_output_port[1]
        calibration_input = 2
        optimizer_parameters = self._get_optimizer_parameters(optimizer_parameters)

        lo_to_if_mapping = create_lo_to_if_list_mapping(lo_if_frequencies_tuple_list)
        if len(lo_to_if_mapping) > 1:  # TODO fix loop bug and remove this error, from here the code assumes 1 lo_freq
            raise NotImplementedError("Multiple lo frequencies is not yet supported.")

        lo_freq, if_freq_list = lo_to_if_mapping.popitem()

        first_if = if_freq_list[0]
        t_end_preparation = perf_counter()
        logger.debug(f"Pre calibration preparation took {t_end_preparation - calibration_start}")
        compiled, qm = self._compile_calibration_program(
            lo_freq,
            first_if,
            octave_output_port,
            optimizer_parameters,
            close_open_quantum_machines,
        )
        t_start_setup = perf_counter()

        state_name_before = self._set_octave_for_calibration(
            octave,
            calibration_input,
            output_port,
            skewed_lo_freq=lo_freq + optimizer_parameters["calibration_offset_frequency"],
        )
        t_end_setup = perf_counter()
        logger.debug(f"Setting octave for LO took {t_end_setup - t_start_setup}.")

        result = {}
        for if_freq in if_freq_list:
            t_start_setup = perf_counter()
            self._set_if_freq(qm, if_freq, optimizer_parameters)
            t_end_setup = perf_counter()
            logger.debug(f"Setting octave for IF took {t_end_setup - t_start_setup}.")

            dc_offsets, correction = _run_compiled(compiled, qm)
            t_end_run = perf_counter()
            logger.debug(f"Running QUA calibration took {t_end_setup - t_end_run}.")
            temp = octave.rf_outputs[output_port].get_temperature()

            result[lo_freq, if_freq] = CalibrationResult(
                mixer_id=octave_output_mixer_name(*octave_output_port),
                correction=correction,
                i_offset=dc_offsets[0],
                q_offset=dc_offsets[1],
                temperature=temp,
                if_frequency=if_freq,
                lo_frequency=lo_freq,
                optimizer_parameters=optimizer_parameters,
            )

            if save_to_db and self._octave_config.calibration_db is not None:
                self._octave_config.calibration_db.update_calibration_data(result[lo_freq, if_freq])

        # set to previous state
        t_start_restoration = perf_counter()
        octave.restore_state(state_name_before)
        # Close the qm
        qm.close()
        t_end = perf_counter()
        logger.debug(f"Restoring setup and closing QM took QUA calibration took {t_end - t_start_restoration}.")
        logger.debug(f"Calibration took {t_end - calibration_start}.")
        return result

    def _get_optimizer_parameters(self, optimizer_parameters):
        optimizer_parameters_defaults = {
            "average_iterations": 100,
            "iterations": 10000,
            "calibration_offset_frequency": 7e6,
            "keep_on": False,
        }
        if optimizer_parameters is not None:
            for p in optimizer_parameters:
                if p in optimizer_parameters_defaults:
                    optimizer_parameters_defaults[p] = optimizer_parameters[p]
                else:
                    raise ValueError(f"optimizer parameter {p} is not supported")
        return optimizer_parameters_defaults

    def _compile_calibration_program(
        self,
        first_lo,
        first_if,
        octave_output_port,
        optimizer_parameters,
        close_open_quantum_machines=True,
    ):
        from qm.octave._calibration_program import _generate_program

        iq_channels = self._octave_config.get_opx_iq_ports(octave_output_port)
        controller_name = iq_channels[0][0]
        adc_channels = [(controller_name, 1), (controller_name, 2)]
        config = _prep_config(
            iq_channels,
            adc_channels,
            first_if,
            first_lo,
            optimizer_parameters,
        )
        prog = _generate_program(optimizer_parameters)
        try:
            t0 = perf_counter()
            qm = self._qmm.open_qm(config, close_other_machines=close_open_quantum_machines)
            logger.debug(f"Creating dedicated QM for calibration took {perf_counter() - t0}")
        except OpenQmException as e:
            raise OpenQmException("Mixer calibration failed: Could not open a quantum machine.") from e

        t0 = perf_counter()
        compiled = qm.compile(prog)
        logger.debug(f"Compiling calibration program took {perf_counter() - t0}.")

        return compiled, qm

    @staticmethod
    def _set_if_freq(qm, if_freq, optimizer_parameters):
        down_mixer_offset, signal_freq, image_freq = _get_frequencies(if_freq, optimizer_parameters)
        with qm.octave.batch_mode():
            qm.set_intermediate_frequency("IQmixer", if_freq)
            qm.set_intermediate_frequency("signal_analyzer", signal_freq)
            qm.set_intermediate_frequency("lo_analyzer", down_mixer_offset)
            qm.set_intermediate_frequency("image_analyzer", image_freq)

    @staticmethod
    def _set_octave_for_calibration(octave, calibration_input, output_port, skewed_lo_freq) -> str:
        octave.start_batch_mode()
        state_name_before = "before_cal"
        octave.snapshot_state(state_name_before)
        # switch to loopback mode to listen in on the RF output
        octave.rf_inputs[calibration_input].set_rf_source(_convert_rf_output_index_to_input_source(output_port))
        octave.rf_inputs[calibration_input].set_if_mode_i(IFMode.direct)
        octave.rf_inputs[calibration_input].set_if_mode_q(IFMode.direct)
        octave.rf_inputs[1].set_if_mode_i(IFMode.off)
        octave.rf_inputs[1].set_if_mode_q(IFMode.off)
        octave.rf_outputs[output_port].set_output(RFOutputMode.on)

        octave.rf_inputs[calibration_input].set_lo_source(RFInputLOSource.Analyzer)

        octave.rf_inputs[calibration_input].set_lo_frequency(RFInputLOSource.Analyzer, skewed_lo_freq)
        octave.end_batch_mode()
        return state_name_before


def create_lo_to_if_list_mapping(lo_if_frequencies_tuple_list: List[Tuple[int, int]]) -> Dict[int, List[int]]:
    lo_to_if_mapping = defaultdict(list)
    for lo_freq, if_freq in lo_if_frequencies_tuple_list:
        lo_to_if_mapping[lo_freq].append(if_freq)
    return lo_to_if_mapping



from typing import TypeVar, Type, Any

T = TypeVar("T")


class Singleton(type):
    _instances = {}

    def __call__(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def delete_instances_for_type(mcs, type_to_delete: Type[T]) -> None:
        if type_to_delete in mcs._instances:
            mcs._instances.pop(type_to_delete)



from octave_sdk import RFInputRFSource, RFOutputMode, IFMode, Octave


RF_OUTPUT_INDEX_TO_INPUT_SOURCE = {
    1: RFInputRFSource.Loopback_RF_out_1,
    2: RFInputRFSource.Loopback_RF_out_2,
    3: RFInputRFSource.Loopback_RF_out_3,
    4: RFInputRFSource.Loopback_RF_out_4,
    5: RFInputRFSource.Loopback_RF_out_5,
}


class OctaveDevice:
    def __init__(self, client: Octave):
        self._client = client

    @property
    def client(self):
        return self._client

    def set_for_calibration(self, calibration_input, output_port):
        state_name_before = "before_cal"
        self._client.snapshot_state(state_name_before)
        # switch to loopback mode to listen in on the RF output
        self._client.rf_inputs[calibration_input].set_rf_source(RF_OUTPUT_INDEX_TO_INPUT_SOURCE[output_port])
        self._client.rf_inputs[calibration_input].set_if_mode_i(IFMode.direct)
        self._client.rf_inputs[calibration_input].set_if_mode_q(IFMode.direct)
        self._client.rf_inputs[1].set_if_mode_i(IFMode.off)
        self._client.rf_inputs[1].set_if_mode_q(IFMode.off)
        self._client.rf_outputs[output_port].set_output(RFOutputMode.on)
        return state_name_before

    def restore_default_state(self):
        self._client.restore_default_state()



from typing import Optional, Union, List
import re

import betterproto
from betterproto.lib.google.protobuf import ListValue, Value

from qm.serialization.expression_serializing_visitor import ExpressionSerializingVisitor
from qm.serialization.qua_node_visitor import QuaNodeVisitor
from qm.exceptions import QmQuaException
from qm.grpc import qua


class QuaSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self._indent = 0
        self._lines = []
        self.tags = []

    def out(self):
        return "\n".join(["from qm.qua import *", ""] + self._lines + [""])

    def _out_lines(self):
        return self._lines

    def _default_enter(self, node) -> bool:
        name = f"{type(node).__module__}.{type(node).__name__}"
        statement = _statements.get(name, None)
        block = _blocks.get(name, None)
        if name not in _dont_print:
            print(f"entering {name}")
        if statement:
            line = statement(node)
            self._line(line)
            if name == "betterproto.lib.google.protobuf.ListValue":
                self._fix_legacy_save(line, node)
        if block:
            self._enter_block(block(node))
        return statement is None

    @staticmethod
    def _get_result_name(value):
        value = value.list_value.values[-1]
        if value.string_value != "":
            return value.string_value
        else:
            return QuaSerializingVisitor._get_result_name(value)

    @staticmethod
    def _search_adc_trace(values):
        ADC_TRACE_NAME = "@macro_adc_trace"
        for value in values:
            if isinstance(value, str):
                if value == ADC_TRACE_NAME:
                    return True
                else:
                    return None
            elif value.string_value == ADC_TRACE_NAME:
                return True
            elif len(value.list_value) > 0:
                return QuaSerializingVisitor._search_adc_trace(value.list_value)
        return False

    @staticmethod
    def _search_auto_added_stream(values: List[Value]):
        is_auto_added_result = isinstance(values[2], Value) and values[2].string_value == "auto"
        return is_auto_added_result

    def _fix_legacy_save(self, sp_line, node: ListValue):
        sp_line = sp_line.split(".")
        trace_name = sp_line[0]
        save_name = sp_line[-1].split('"')[1]
        save_name = re.sub(r"_input\d", "", save_name)
        is_auto_added_stream = self._search_auto_added_stream(node.values)
        if is_auto_added_stream:
            self._lines.pop()
            line_to_remove_index = None
            for i in range(len(self._lines)):
                if self._lines[i].find(f"{trace_name} = declare_stream(adc_trace=True)") > 0:
                    line_to_remove_index = i
                self._lines[i] = self._lines[i].replace(trace_name, f'"{save_name}"')
            if line_to_remove_index:
                self._lines.pop(line_to_remove_index)

    def enter_qm_grpc_qua_QuaResultAnalysis(self, node: qua.QuaResultAnalysis) -> bool:
        if len(node.model) > 0:
            self._enter_block("with stream_processing():")
            return True
        else:
            return False

    def enter_qm_grpc_qua_QuaProgramSaveStatement(self, node: qua.QuaProgramSaveStatement) -> bool:
        if node.tag not in self.tags:
            self._line(f"{node.tag} = declare_stream()")
            self.tags.append(node.tag)
        save_line = f"save({ExpressionSerializingVisitor.serialize(node.source)}, {node.tag})"
        self._line(save_line)
        return False

    def enter_qm_grpc_qua_QuaProgramVarDeclaration(self, node: qua.QuaProgramVarDeclaration):
        args = self._get_declare_var_args(node)

        if node.is_input_stream:
            # Removes the '_input_stream' from the end of the name
            stream_name = node.name[13:]
            self._line(
                f"{node.name} = declare_input_stream({_var_type_dec[node.type]}, '{stream_name}', {_dict_to_python_call(args)})"
            )
        else:
            self._line(f"{node.name} = declare({_var_type_dec[node.type]}, {_dict_to_python_call(args)})")

        return False

    def _get_declare_var_args(self, node):
        size = node.size
        dim = node.dim
        args = {}
        if dim > 0:
            args["size"] = size
        if dim > 0 and len(node.value) > 0:
            args["value"] = "[" + ", ".join([_ser_exp(it) for it in node.value]) + "]"
        elif len(node.value) == 1:
            args["value"] = _ser_exp(node.value[0])
        if "value" in args and "size" in args:
            del args["size"]
        return args

    def enter_qm_grpc_qua_QuaProgramMeasureStatement(self, node: qua.QuaProgramMeasureStatement):
        if node.timestamp_label and node.timestamp_label not in self.tags:
            self._line(f"{node.timestamp_label} = declare_stream()")
            self.tags.append(node.timestamp_label)
        if node.stream_as and node.stream_as not in self.tags:
            if node.stream_as.startswith("atr_"):
                self._line(f"{node.stream_as} = declare_stream(adc_trace=True)")
            else:
                self._line(f"{node.stream_as} = declare_stream()")
            self.tags.append(node.stream_as)
        return self._default_enter(node)

    def visit_qm_grpc_qua_QuaProgramForStatement(self, node: qua.QuaProgramForStatement):
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_qm_grpc_qua_QuaProgramForEachStatement(self, node: qua.QuaProgramForEachStatement):
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_qm_grpc_qua_QuaProgramIfStatement(self, node: qua.QuaProgramIfStatement):
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")
        elseifs_list = node.elseifs
        for elseif in elseifs_list:
            elseif_block = elseif.body
            condition = elseif.condition
            self._leave_block()
            self._line(f"with elif_({ExpressionSerializingVisitor.serialize(condition)}):")
            self._enter_block()
            if len(elseif_block.statements) > 0:
                super()._default_visit(elseif_block)
            else:
                self._line("pass")

        else_block = node.else_
        if len(else_block.statements) > 0:  # Cannot ID else_() with pass
            self._leave_block()
            self._line("with else_():")
            self._enter_block()
            super()._default_visit(else_block)

    def visit_qm_grpc_qua_QuaProgramPlayStatement(self, node: qua.QuaProgramPlayStatement):
        pulse_one_of, value = betterproto.which_one_of(node, "pulseType")
        if pulse_one_of == "named_pulse":
            pulse = f'"{node.named_pulse.name}"'
            # node.pulse - duplicate with namedPulse
        elif pulse_one_of == "ramp_pulse":
            pulse = f"ramp({ExpressionSerializingVisitor.serialize(node.ramp_pulse)})"
        else:
            raise QmQuaException(f"Unknown pulse type {pulse_one_of}")

        element = node.qe.name
        amp = ""
        if betterproto.serialized_on_wire(node.amp):
            v0 = _ser_exp(node.amp.v0)
            v1 = _ser_exp(node.amp.v1)
            v2 = _ser_exp(node.amp.v2)
            v3 = _ser_exp(node.amp.v3)
            if v0 != "":
                if v1 != "":
                    amp = f"*amp({v0}, {v1}, {v2}, {v3})"
                else:
                    amp = f"*amp({v0})"
        args = []

        _, duration = betterproto.which_one_of(node.duration, "expression_oneof")
        if duration is not None:
            args.append(f"duration={_ser_exp(duration)}")

        _, condition = betterproto.which_one_of(node.condition, "expression_oneof")
        if condition is not None:
            args.append(f"condition={_ser_exp(condition)}")

        if node.target_input:
            args.append(f'target="{node.target_input}"')

        if betterproto.serialized_on_wire(node.chirp):
            rate_one_of, value = betterproto.which_one_of(node.chirp, "rate")
            if rate_one_of in ("scalar_rate", "array_rate"):
                rate = ExpressionSerializingVisitor.serialize(value)
            else:
                raise QmQuaException(f"Unknown chirp rate {rate_one_of}")

            unit_mapping = {
                qua.QuaProgramChirpUnits.HzPerNanoSec: "Hz/nsec",
                qua.QuaProgramChirpUnits.mHzPerNanoSec: "mHz/nsec",
                qua.QuaProgramChirpUnits.uHzPerNanoSec: "uHz/nsec",
                qua.QuaProgramChirpUnits.nHzPerNanoSec: "nHz/nsec",
                qua.QuaProgramChirpUnits.pHzPerNanoSec: "pHz/nsec",
            }
            s_units = node.chirp.units
            if s_units in unit_mapping:
                units = unit_mapping[s_units]
            else:
                raise QmQuaException(f"Unsupported units {s_units}")

            times_builder = []
            for time in node.chirp.times:
                times_builder.append(f"{time}")

            if len(times_builder) > 0:
                times = f'[{", ".join(times_builder)}]'
            else:
                times = "None"

            if node.chirp.continue_chirp:
                args.append("continue_chirp=True")

            args.append(f'chirp=({rate},{times},"{units}")')

        _, truncate = betterproto.which_one_of(node.truncate, "expression_oneof")
        if truncate is not None:
            args.append(f"truncate={_ser_exp(truncate)}")

        if node.timestamp_label:
            if node.timestamp_label not in self.tags:
                self._line(f"{node.timestamp_label} = declare_stream()")
                self.tags.append(node.timestamp_label)
            args.append(f"timestamp_stream={node.timestamp_label}")

        # TODO maybe make sure no other fields?

        if len(args) > 0:
            args_str = f', {", ".join(args)}'
        else:
            args_str = ""
        self._line(f'play({pulse}{amp}, "{element}"{args_str})')

    def _default_leave(self, node):
        name = f"{type(node).__module__}.{type(node).__name__}"
        block = _blocks.get(name, None)
        if block is not None:
            self._leave_block()
        super()._default_leave(node)

    def _enter_block(self, line: Optional[str] = None):
        if line is not None:
            self._line(line)
        self._indent += 1

    def _leave_block(self):
        self._indent -= 1

    def _line(self, line: str):
        self._lines.append((self._indent * "    ") + line)


def _ser_exp(value):
    return ExpressionSerializingVisitor.serialize(value)


def _dict_to_python_call(d: dict) -> str:
    return ", ".join([f"{k}={v}" for k, v in d.items()])


def _ramp_to_zero_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    if node.duration is not None:
        args.append(str(node.duration))
    return f'ramp_to_zero({", ".join(args)})'


def _measure_statement(node: qua.QuaProgramMeasureStatement):
    args = []

    amp = ""
    v0 = _ser_exp(node.amp.v0)
    v1 = _ser_exp(node.amp.v1)
    v2 = _ser_exp(node.amp.v2)
    v3 = _ser_exp(node.amp.v3)
    if v0 != "":
        if v1 != "":
            amp = f"*amp({v0}, {v1}, {v2}, {v3})"
        else:
            amp = f"*amp({v0})"

    args.append(f'"{node.pulse.name}"{amp}')
    args.append(f'"{node.qe.name}"')
    if node.stream_as:
        args.append(f"{node.stream_as}")
    else:
        args.append("None")

    if len(node.measure_processes) > 0:
        for process in node.measure_processes:
            args.append(ExpressionSerializingVisitor.serialize(process))
    if node.timestamp_label:
        args.append(f"timestamp_stream={node.timestamp_label}")
    return f'measure({", ".join(args)})'


def _wait_statement(node: qua.QuaProgramWaitStatement):
    args = []
    _, wait_value = betterproto.which_one_of(node.time, "expression_oneof")
    if wait_value is not None:
        args.append(f"{ExpressionSerializingVisitor.serialize(wait_value)}")
    qes = []
    for qe in node.qe:
        qes.append(f'"{qe.name}"')
    args.append(", ".join(qes))
    return f'wait({", ".join(args)})'


def _align_statement(node):
    args = []
    for qe in node.qe:
        args.append(f'"{qe.name}"')
    return f'align({", ".join(args)})'


def _wait_for_trigger_statement(node: qua.QuaProgramWaitForTriggerStatement):
    args = []
    for qe in node.qe:
        args.append(f'"{qe.name}"')
    if node.pulse_to_play.name:
        args.append(f'"{node.pulse_to_play.name}"')
    if node.element_output.element:
        if node.element_output.output:
            args.append(f'trigger_element=("{node.element_output.element}", "{node.element_output.output}")')
        else:
            args.append(f'trigger_element="{node.element_output.element}"')
    if node.time_tag_target.name != "":
        args.append(f"time_tag_target={node.time_tag_target.name}")
    return f'wait_for_trigger({", ".join(args)})'


def _frame_rotation_statement(node):
    args = []
    args.append(f"{ExpressionSerializingVisitor.serialize(node.value)}")
    args.append(f'"{node.qe.name}"')
    return f'frame_rotation_2pi({", ".join(args)})'


def _fast_frame_rotation_statement(node):
    args = []
    args.append(f"{ExpressionSerializingVisitor.serialize(node.cosine)}")
    args.append(f"{ExpressionSerializingVisitor.serialize(node.sine)}")
    args.append(f'"{node.qe.name}"')
    return f'fast_frame_rotation({", ".join(args)})'


def _reset_frame_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    return f'reset_frame({", ".join(args)})'


def _update_frequency_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    args.append(f"{ExpressionSerializingVisitor.serialize(node.value)}")
    args.append(f'"{qua.QuaProgramUpdateFrequencyStatementUnits(node.units).name}"')
    args.append(f"{node.keep_phase}")
    return f'update_frequency({", ".join(args)})'


def _set_dc_offset_statement(node: qua.QuaProgramSetDcOffsetStatement):
    args = []
    args.append(f'"{node.qe.name}"')
    args.append(f'"{node.qe_input_reference}"')
    args.append(f"{ExpressionSerializingVisitor.serialize(node.offset)}")
    return f'set_dc_offset({", ".join(args)})'


def _advance_input_stream_statement(node):
    if node.stream_array.name != "":
        input_stream = ExpressionSerializingVisitor.serialize(node.stream_array)
    elif node.stream_variable.name != "":
        input_stream = ExpressionSerializingVisitor.serialize(node.stream_variable)
    else:
        raise RuntimeError("unsupported type for pop input stream")
    return f"advance_input_stream({input_stream})"


def _for_block_statement(node):
    condition = ExpressionSerializingVisitor.serialize(node.condition)
    if len(node.init.statements) == 0 and len(node.update.statements) == 0:
        if condition == "True":
            return "with infinite_loop_():"
        else:
            return f"with while_({condition}):"
    else:
        if len(node.init.statements) != 1:
            raise Exception("for is not valid")
        if len(node.update.statements) != 1:
            raise Exception("for is not valid")

        init = node.init.statements[0].assign
        update = node.update.statements[0].assign

        if init is None:
            raise Exception("for is not valid")
        if update is None:
            raise Exception("for is not valid")

        return f"with for_({ExpressionSerializingVisitor.serialize(init.target)},{ExpressionSerializingVisitor.serialize(init.expression)},{condition},{ExpressionSerializingVisitor.serialize(update.expression)}):"


def _for_each_block_statement(node):
    variables = []
    arrays = []
    for it in node.iterator:
        variables.append(_ser_exp(it.variable))
        arrays.append(it.array.name)
    return f'with for_each_(({",".join(variables)}),({",".join(arrays)})):'


def _if_block_statement(node):
    condition = ExpressionSerializingVisitor.serialize(node.condition)
    if node.unsafe is True:
        unsafe = ", unsafe=True"
    else:
        unsafe = ""
    return f"with if_({condition}{unsafe}):"


def _strict_timing_block_statement(_):
    return "with strict_timing_():"


def _stream_processing_function(array):
    function = array[0].string_value

    if function == "average":
        if len(array) > 1:
            var = array[1].string_value
        else:
            var = ""
        return f"average({var})"

    if function == "dot":
        if len(array) == 1:
            return "tuple_dot_product()"
        else:
            vector = _stream_processing_operator(array[1].list_value.values)
            return f"dot_product({vector})"

    if function == "vmult":
        vector = _stream_processing_operator(array[1].list_value.values)
        return f"multiply_by({vector})"

    if function == "smult":
        return f"multiply_by({array[1].string_value})"

    if function == "tmult":
        return "tuple_multiply()"

    if function == "conv":
        if len(array) == 2:
            if array[1].string_value:
                mode = f'"{array[1].string_value}"'
            else:
                mode = ""
            return f"tuple_convolution({mode})"
        else:
            vector = _stream_processing_operator(array[2].list_value.values)
            if array[1].string_value:
                mode = f',"{array[1].string_value}"'
            else:
                mode = ""
            return f"convolution({vector}{mode})"

    if function == "fft":
        return "fft()"

    if function == "booleancast":
        return "boolean_to_int()"

    if function == "demod":
        if len(array[2].list_value) > 0:
            cos = _stream_processing_operator(array[2].list_value.values)
        else:
            cos = array[2].string_value
        if len(array[3].list_value) > 0:
            sin = _stream_processing_operator(array[3].list_value.values)
        else:
            sin = array[3].string_value
        return f"demod({array[1].string_value},{cos},{sin})"

    print(f"missing function: {function}")
    return "default_function()"


def _stream_processing_operator(array: List[Value]) -> str:
    operator = array[0].string_value

    if operator == "@macro_input":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.input{array[1].string_value}()"

    if operator == "@array":
        values = []
        for a in array[1:]:
            value = _stream_processing_statement(a)
            values.append(value)
        return f'[{", ".join(values)}]'

    if operator == "@macro_auto_reshape":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.auto_reshape()"

    if operator == "+":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.add({right})"  # arithmetic stream

    if operator == "-":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.subtract({right})"  # arithmetic stream

    if operator == "/":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.divide({right})"  # arithmetic stream

    if operator == "*":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.multiply({right})"  # arithmetic stream

    if operator == "zip":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{right}.zip({left})"  # arithmetic stream

    if operator == "take":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.take({array[1].string_value})"

    if operator == "buffer":
        chain = _default_stream_processing_chain(array)
        dims = [dim.string_value for dim in array[1:-1]]
        return f'{chain}.buffer({", ".join(dims)})'

    if operator == "bufferAndSkip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.buffer_and_skip({array[1].string_value}, {array[2].string_value})"

    if operator == "skip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip({array[1].string_value})"

    if operator == "skipLast":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip_last({array[1].string_value})"

    if operator == "histogram":
        chain = _default_stream_processing_chain(array)
        bins = _stream_processing_operator(array[1].list_value.values)
        return f"{chain}.histogram({bins})"

    if operator == "average":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.average()"

    if operator == "flatten":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.flatten()"

    if operator == "map":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.map(FUNCTIONS.{_stream_processing_function(array[1].list_value.values)})"

    if operator == "@re":
        chain = _default_stream_processing_chain(array)
        timestamp_mode = int(array[1].string_value)
        if timestamp_mode == 0:  # values
            return f"{chain}"
        elif timestamp_mode == 1:  # timestamps
            return f"{chain}.timestamps()"
        elif timestamp_mode == 2:  # ValuesAndTimestamps
            return f"{chain}.with_timestamps()"

    if operator == "@macro_adc_trace":
        chain = _default_stream_processing_chain(array)
        return f"{chain}"

    print(f"missing operator: {operator}")
    chain = _default_stream_processing_chain(array)
    return f"{chain}"


def _default_stream_processing_chain(array: List[Value]) -> str:
    last_index = len(array) - 1
    chain = _stream_processing_statement(array[last_index])
    return f"{chain}"


def _stream_processing_statement(node: Union[ListValue, Value]) -> str:
    if len(node.list_value.values) > 0:
        return _stream_processing_operator(node.list_value.values)
    else:
        return node.string_value


def _stream_processing_terminal_statement(node: ListValue):
    last_index = len(node.values) - 1
    chain = _stream_processing_statement(node.values[last_index])
    terminal = node.values[0].string_value
    terminal = "save_all" if terminal == "saveAll" else terminal  # normalize save all
    return f'{chain}.{terminal}("{node.values[1].string_value}")'


def _serialize(node):
    visitor = QuaSerializingVisitor()
    visitor.visit(node)
    return visitor._out_lines()


_blocks = {
    "qm.grpc.qua.QuaProgram": lambda n: "with program() as prog:",
    "qm.grpc.qua.QuaProgramForStatement": _for_block_statement,
    "qm.grpc.qua.QuaProgramForEachStatement": _for_each_block_statement,
    "qm.grpc.qua.QuaProgramIfStatement": _if_block_statement,
    "qm.grpc.qua.QuaProgramStrictTimingStatement": _strict_timing_block_statement,
}

_statements = {
    "qm.grpc.qua.QuaProgramMeasureStatement": _measure_statement,
    "qm.grpc.qua.QuaProgramWaitStatement": _wait_statement,
    "qm.grpc.qua.QuaProgramAssignmentStatement": lambda n: f"assign({ExpressionSerializingVisitor.serialize(n.target)}, "
    f"{ExpressionSerializingVisitor.serialize(n.expression)})",
    "qm.grpc.qua.QuaProgramPauseStatement": lambda n: "pause()",
    "qm.grpc.qua.QuaProgramResetPhaseStatement": lambda n: f'reset_phase("{n.qe.name}")',
    "qm.grpc.qua.QuaProgramUpdateFrequencyStatement": _update_frequency_statement,
    "qm.grpc.qua.QuaProgramAlignStatement": _align_statement,
    "qm.grpc.qua.QuaProgramWaitForTriggerStatement": _wait_for_trigger_statement,
    "qm.grpc.qua.QuaProgramZRotationStatement": _frame_rotation_statement,
    "qm.grpc.qua.QuaProgramRampToZeroStatement": _ramp_to_zero_statement,
    "qm.grpc.qua.QuaProgramResetFrameStatement": _reset_frame_statement,
    "betterproto.lib.google.protobuf.ListValue": _stream_processing_terminal_statement,
    # list value statement is assumed just as stream processing for now
    "qm.grpc.qua.QuaProgramUpdateCorrectionStatement": lambda n: f'update_correction("{n.qe.name}"'
    f",{ExpressionSerializingVisitor.serialize(n.correction.c0)}"
    f",{ExpressionSerializingVisitor.serialize(n.correction.c1)}"
    f",{ExpressionSerializingVisitor.serialize(n.correction.c2)}"
    f",{ExpressionSerializingVisitor.serialize(n.correction.c3)})",
    "qm.grpc.qua.QuaProgramSetDcOffsetStatement": _set_dc_offset_statement,
    "qm.grpc.qua.QuaProgramAdvanceInputStreamStatement": _advance_input_stream_statement,
    "qm.grpc.qua.QuaProgramFastFrameRotationStatement": _fast_frame_rotation_statement,
}

_var_type_dec = {0: "int", 1: "bool", 2: "fixed"}

_nodes_to_ignore = [
    "qm.grpc.qua.QuaProgramScript",
    "qm.grpc.qua.QuaProgramStatementsCollection",
    "qm.grpc.qua.QuaProgramAnyStatement",
    "qm.grpc.qua.QuaProgramAnyScalarExpression",
    "qm.grpc.qua.QuaProgramBinaryExpression",
    "qm.grpc.qua.QuaProgramVarRefExpression",
    "qm.grpc.qua.QuaProgramLiteralExpression",
    "qm.grpc.qua.QuaProgramPlayStatement",
    "qm.grpc.qua_config.QuaConfig",
]

_dont_print = set(list(_blocks.keys()) + list(_statements.keys()) + _nodes_to_ignore)



import json as _json
import warnings
from dataclasses import dataclass
from io import BytesIO
from typing import Dict, BinaryIO, Any, Optional, Callable

import betterproto
import numpy
import numpy.typing
from betterproto.lib.google.protobuf import Value, Struct
from numpy.lib import format as _format

from qm.grpc.results_analyser import SimulatorSamplesResponseHeader, SimulatorSamplesResponseData
from qm.type_hinting.simulator_types import AnalogOutputsType, DigitalOutputsType, WaveformInPortsType
from qm.utils import deprecation_message
from qm.utils.async_utils import run_async
from qm.api.frontend_api import FrontendApi
from qm.api.models.capabilities import ServerCapabilities
from qm.api.simulation_api import SimulationApi
from qm.exceptions import QMSimulationError
from qm.grpc.frontend import SimulatedResponsePart
from qm.jobs.running_qm_job import RunningQmJob
from qm.persistence import BaseStore
from qm.results.simulator_samples import SimulatorSamples
from qm.waveform_report import WaveformReport


@dataclass
class DensityMatrix:
    timestamp: int
    data: numpy.typing.NDArray[numpy.cdouble]


class SimulatorOutput:
    def __init__(self, job_id: str, frontend: FrontendApi) -> None:
        super().__init__()
        self._id = job_id
        self._simulation_api = SimulationApi.from_api(frontend)

    def get_quantum_state(self) -> DensityMatrix:

        state = self._simulation_api.get_simulated_quantum_state(self._id)
        flatten = numpy.array([complex(item.re, item.im) for item in state.data])
        n = int(numpy.sqrt(len(flatten)))
        if len(flatten) != (n * n):
            raise RuntimeError("Quantum state matrix is not correct")
        else:
            pass

        matrix = flatten.reshape(n, n)
        timestamp = state.time_stamp
        return DensityMatrix(timestamp, matrix)


def extract_value(value: Value) -> Any:
    name, one_of = betterproto.which_one_of(value, "kind")
    if name in VALUE_MAPPING:
        return VALUE_MAPPING[name](one_of)


def extract_struct_value(struct_value: Struct) -> Any:
    output = {}
    for name, value in struct_value.fields.items():
        output[name] = extract_value(value)
    return output


VALUE_MAPPING: Dict[str, Callable[[Any], Any]] = {
    "number_value": float,
    "string_value": str,
    "bool_value": bool,
    "list_value": lambda list_value: [extract_value(value) for value in list_value.values],
    "struct_value": extract_struct_value,
    "null_value": lambda x: None,
}


class SimulatedJob(RunningQmJob):
    def __init__(
        self,
        job_id: str,
        frontend_api: FrontendApi,
        capabilities: ServerCapabilities,
        store: BaseStore,
        simulated_response: SimulatedResponsePart,
    ):
        super().__init__(job_id, "", frontend_api, capabilities, store)
        self._waveform_report: Optional[WaveformReport] = None

        self._simulated_analog_outputs: AnalogOutputsType = {"waveforms": None}
        self._simulated_digital_outputs: DigitalOutputsType = {"waveforms": None}
        if betterproto.serialized_on_wire(simulated_response.analog_outputs):
            self._simulated_analog_outputs = extract_struct_value(simulated_response.analog_outputs)
        if betterproto.serialized_on_wire(simulated_response.digital_outputs):
            self._simulated_digital_outputs = extract_struct_value(simulated_response.digital_outputs)
        if betterproto.serialized_on_wire(simulated_response.waveform_report):
            self._waveform_report = WaveformReport.from_dict(
                extract_struct_value(simulated_response.waveform_report), self.id
            )
        if simulated_response.errors:
            raise QMSimulationError("\n".join(simulated_response.errors))

        self._simulation_api = SimulationApi.from_api(self._frontend)

    def _initialize_from_job_status(self) -> None:
        """
        Overriding this for simulated jobs to do nothing
        """
        pass

    def get_simulated_waveform_report(self) -> Optional[WaveformReport]:
        return self._waveform_report

    def simulated_analog_waveforms(self) -> Optional[WaveformInPortsType]:
        """
        Return the results of the simulation of elements and analog outputs.

        The returned dictionary has the following keys and entries:

        - **elements**: a dictionary containing the outputs with timestamps and values arranged by elements.

        - **controllers: a dictionary containing the outputs with timestamps and values arranged by controllers.

            - **ports**: a dictionary containing the outputs with timestamps and values arranged by output ports.
                for each element or output port, the entry is a list of dictionaries with the following information:

                - **timestamp**:
                    The time, in nsec, from the start of the program to the start of the pulse.

                - **samples**:

                    Output information, with ``duration`` given in nsec and ``value`` given normalized OPX output units.

        Returns:
            A dictionary containing output information for the analog outputs of the controller.

        """
        warnings.warn(
            deprecation_message(
                method="SimulatedJob.simulated_analog_waveforms",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use 'get_simulated_waveform_report' instead.",
            ),
            DeprecationWarning,
        )
        return self._simulated_analog_outputs["waveforms"]

    def simulated_digital_waveforms(self) -> Optional[WaveformInPortsType]:
        """
        Return the results of the simulation of digital outputs.

        - **controllers**: a dictionary containing the outputs with timestamps and values arranged by controllers.

            - **ports**: a dictionary containing the outputs with timestamps and values arranged by output ports.
                for each element or output port, the entry is a list of dictionaries with the following information:

                - **timestamp**:
                    The time, in nsec, from the start of the program to the start of the pulse.

                - **samples**:
                    A list containing the sequence of outputted values, with ``duration`` given in nsec
                    and ``value`` given as a boolean value.

        Returns:
            A dictionary containing output information for the analog outputs of the controller.
        """
        warnings.warn(
            deprecation_message(
                method="SimulatedJob.simulated_digital_waveforms",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use 'get_simulated_waveform_report' instead.",
            ),
            DeprecationWarning,
        )
        return self._simulated_digital_outputs["waveforms"]

    async def _pull_simulator_samples(
        self, include_analog: bool, include_digital: bool, writer: BinaryIO, data_writer: BinaryIO
    ) -> None:
        async for result in self._simulation_api.pull_simulator_samples(self._id, include_analog, include_digital):
            if result.ok:
                name, value = betterproto.which_one_of(result, "body")
                if name == "header" and isinstance(value, SimulatorSamplesResponseHeader):
                    _format.write_array_header_2_0(  # type: ignore[no-untyped-call]
                        writer,
                        {
                            "descr": _json.loads(value.simple_d_type),
                            "fortran_order": False,
                            "shape": (value.count_of_items,),
                        },
                    )
                elif name == "data" and isinstance(value, SimulatorSamplesResponseData):
                    data_writer.write(value.data)
            else:
                raise QMSimulationError("Error while pulling samples")

    def _get_np_simulated_samples(
        self, include_analog: bool = True, include_digital: bool = True
    ) -> numpy.typing.NDArray[numpy.generic]:
        writer = BytesIO()
        data_writer = BytesIO()

        run_async(self._pull_simulator_samples(include_analog, include_digital, writer, data_writer))

        data_writer.seek(0)
        for d in data_writer:
            writer.write(d)

        writer.seek(0)
        ret: numpy.typing.NDArray[numpy.generic] = numpy.load(writer)  # type: ignore[no-untyped-call]
        return ret

    def get_simulated_samples(self, include_analog: bool = True, include_digital: bool = True) -> SimulatorSamples:
        """
        Obtain the output samples of a QUA program simulation.

        Samples are returned in an object that holds the controllers in the current simulation,
        where each controller's name will be a property of this object.
        The value of each property of the returned value is an object with the following properties:

        ``analog``:

            holds a dictionary with analog port names as keys and numpy array of samples as values.

        ``digital``:

            holds a dictionary with digital port names as keys and numpy array of samples as values.

        It is also possible to directly plot the outputs using a built-in plot command.

        Example:
            ```python
            samples = job.get_simulated_samples()
            analog1 = samples.con1.analog["1"]  # obtain analog port 1 of controller "con1"
            digital9 = samples.con1.analog["9"]  # obtain digital port 9 of controller "con1"
            samples.con1.plot()  # Plots all active ports
            samples.con1.plot(analog_ports=['1', '2'], digital_ports=['9'])  # Plots the given output ports
            ```

        !!! Note

            The simulated digital waveforms are delayed by 136ns relative to the real
            output of the OPX.

        Args:
            include_analog: Should we collect simulated analog samples
            include_digital: Should we collect simulated digital samples

        Returns:
            The simulated samples of the job.
        """
        return SimulatorSamples.from_np_array(
            self._get_np_simulated_samples(include_analog=include_analog, include_digital=include_digital)
        )

    @property
    def simulator(self) -> SimulatorOutput:
        return SimulatorOutput(self._id, self._frontend)



from typing import Dict, Union, Tuple, List
from typing_extensions import TypedDict, Literal

from qm.type_hinting.general import Number

PortReferenceType = Tuple[str, int]


# TODO: This is a placeholder while we still use dicts, once we move to pydantics we can simply change the
#  inheritance of the classes handled here and add a more robust validation to the types


class AnalogOutputFilterConfigType(TypedDict, total=False):
    feedforward: List[float]
    feedback: List[float]


class AnalogOutputPortConfigType(TypedDict, total=False):
    offset: Number
    filter: AnalogOutputFilterConfigType
    delay: int
    crosstalk: Dict[int, Number]
    shareable: bool
    connectivity: Tuple[str, str]


class AnalogInputPortConfigType(TypedDict, total=False):
    offset: Number
    gain_db: int
    shareable: bool
    connectivity: Tuple[str, str]


class DigitalOutputPortConfigType(TypedDict, total=False):
    sharable: bool
    inverted: bool
    connectivity: Tuple[str, str]


class DigitalInputPortConfigType(TypedDict, total=False):
    deadtime: int
    polarity: str
    threshold: Number
    connectivity: Tuple[str, str]


class OctaveConfigType(TypedDict, total=False):
    loopbacks: List[Tuple[Tuple[str, str], str]]


class ControllerConfigType(TypedDict, total=False):
    type: Literal["opx1"]
    analog_outputs: Dict[int, AnalogOutputPortConfigType]
    analog_inputs: Dict[int, AnalogInputPortConfigType]
    digital_outputs: Dict[int, DigitalOutputPortConfigType]
    digital_inputs: Dict[int, DigitalInputPortConfigType]
    connectivity: str


class DigitalInputConfigType(TypedDict, total=False):
    delay: int
    buffer: int
    port: PortReferenceType


class IntegrationWeightConfigType(TypedDict, total=False):
    cosine: Union[List[Tuple[float, int]], List[float]]
    sine: Union[List[Tuple[float, int]], List[float]]


class ConstantWaveFormConfigType(TypedDict, total=False):
    type: str
    sample: float


class ArbitraryWaveFormConfigType(TypedDict, total=False):
    type: str
    samples: List[float]
    max_allowed_error: float
    sampling_rate: Number
    is_overridable: bool


class DigitalWaveformConfigType(TypedDict, total=False):
    samples: List[Tuple[int, int]]


class MixerConfigType(TypedDict, total=False):
    intermediate_frequency: float
    lo_frequency: float
    correction: Tuple[Number, Number, Number, Number]


class PulseConfigType(TypedDict, total=False):
    operation: str
    length: int
    waveforms: Dict[str, str]
    digital_marker: str
    integration_weights: Dict[str, str]


class SingleInputConfigType(TypedDict, total=False):
    port: PortReferenceType


class HoldOffsetConfigType(TypedDict, total=False):
    duration: int


class MixInputOctaveParamsConfigType(TypedDict, total=False):
    rf_output_port: Tuple[str, str]
    lo_source: str
    output_switch_state: str
    output_gain: int
    downconversion_lo_source: str
    downconversion_lo_frequency: float
    rf_input_port: Tuple[str, str]


class StickyConfigType(TypedDict, total=False):
    analog: bool
    digital: bool
    duration: int


class MixInputConfigType(TypedDict, total=False):
    I: PortReferenceType
    Q: PortReferenceType
    mixer: str
    lo_frequency: float
    octave_params: MixInputOctaveParamsConfigType


class InputCollectionConfigType(TypedDict, total=False):
    inputs: Dict[str, PortReferenceType]


class OscillatorConfigType(TypedDict, total=False):
    intermediate_frequency: float
    mixer: str
    lo_frequency: float


class OutputPulseParameterConfigType(TypedDict):
    signalThreshold: int
    signalPolarity: str
    derivativeThreshold: int
    derivativePolarity: str


class ElementConfigType(TypedDict, total=False):
    intermediate_frequency: float
    oscillator: str
    measurement_qe: str
    operations: Dict[str, str]
    singleInput: SingleInputConfigType
    mixInputs: MixInputConfigType
    singleInputCollection: InputCollectionConfigType
    multipleInputs: InputCollectionConfigType
    time_of_flight: int
    smearing: int
    outputs: Dict[str, PortReferenceType]
    digitalInputs: Dict[str, DigitalInputConfigType]
    digitalOutputs: Dict[str, PortReferenceType]
    outputPulseParameters: OutputPulseParameterConfigType
    hold_offset: HoldOffsetConfigType
    sticky: StickyConfigType
    thread: str


class DictQuaConfig(TypedDict, total=False):
    version: int
    oscillators: Dict[str, OscillatorConfigType]
    elements: Dict[str, ElementConfigType]
    controllers: Dict[str, ControllerConfigType]
    octaves: Dict[str, OctaveConfigType]
    integration_weights: Dict[str, IntegrationWeightConfigType]
    waveforms: Dict[str, Union[ArbitraryWaveFormConfigType, ConstantWaveFormConfigType]]
    digital_waveforms: Dict[str, DigitalWaveformConfigType]
    pulses: Dict[str, PulseConfigType]
    mixers: Dict[str, List[MixerConfigType]]



import dataclasses
from typing import Optional, List

from qm.grpc.qua import (
    QuaProgram,
    QuaProgramType,
    QuaProgramLiteralExpression,
    QuaProgramVarDeclaration,
    QuaProgramScript,
    QuaResultAnalysis,
    QuaProgramStatementsCollection,
)
from qm.grpc.qua_config import QuaConfig
from qm.program.StatementsCollection import StatementsCollection
from qm.program._ResultAnalysis import _ResultAnalysis


@dataclasses.dataclass
class ProgramMetadata:
    uses_command_timestamps: bool
    uses_fast_frame_rotation: bool


class Program:
    def __init__(
        self,
        config: Optional[QuaConfig] = None,
        program: Optional[QuaProgram] = None,
    ):
        if program is None:
            program = QuaProgram(
                script=QuaProgramScript(variables=[], body=QuaProgramStatementsCollection(statements=[])),
                result_analysis=QuaResultAnalysis(model=[]),
            )

        self._program = program
        self._qua_config = config
        self._result_analysis = _ResultAnalysis(self._program.result_analysis)
        self._is_in_scope = False
        self._metadata = ProgramMetadata(uses_command_timestamps=False, uses_fast_frame_rotation=False)

    def _declare_var(
        self,
        name: str,
        var_type: QuaProgramType,
        size: int,
        value: Optional[List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ):
        declaration = QuaProgramVarDeclaration(
            name=name,
            type=var_type,
            size=size,
            dim=dim,
            is_input_stream=is_input_stream,
        )

        if value is None:
            pass
        elif type(value) is list:
            for i in value:
                declaration.value.append(i)
        else:
            declaration.value.append(value)
        self._program.script.variables.append(declaration)

    def declare_int(
        self,
        name: str,
        size: int,
        value: Optional[List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ):
        self._declare_var(name, QuaProgramType.INT, size, value, dim, is_input_stream)

    def declare_real(
        self,
        name: str,
        size: int,
        value: Optional[List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ):
        self._declare_var(name, QuaProgramType.REAL, size, value, dim, is_input_stream)

    def declare_bool(
        self,
        name: str,
        size: int,
        value: Optional[List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ):
        self._declare_var(name, QuaProgramType.BOOL, size, value, dim, is_input_stream)

    @property
    def body(self) -> StatementsCollection:
        return StatementsCollection(self._program.script.body)

    @property
    def result_analysis(self) -> _ResultAnalysis:
        return self._result_analysis

    def build(self, config: QuaConfig) -> QuaProgram:
        copy = QuaProgram().from_dict(self._program.to_dict())
        copy.config = QuaConfig().from_dict(config.to_dict())
        return copy

    def set_in_scope(self):
        self._is_in_scope = True

    def set_exit_scope(self):
        self._is_in_scope = False

    def is_in_scope(self) -> bool:
        return self._is_in_scope

    @property
    def metadata(self) -> ProgramMetadata:
        return self._metadata

    def set_metadata(
        self,
        uses_command_timestamps: bool = None,
        uses_fast_frame_rotation: bool = None,
    ):
        if uses_command_timestamps is not None:
            self._metadata.uses_command_timestamps = uses_command_timestamps
        if uses_fast_frame_rotation is not None:
            self.metadata.uses_fast_frame_rotation = uses_fast_frame_rotation



from qm.results import (
    StreamingResultFetcher,
    SingleStreamingResultFetcher,
    MultipleStreamingResultFetcher,
)  # noqa
import warnings

# Backward compatible names
JobResults = StreamingResultFetcher
SingleNamedJobResult = SingleStreamingResultFetcher
MultipleNamedJobResult = MultipleStreamingResultFetcher

warnings.warn(
    "'qm._results.JobResults' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.results.JobResults' instead",
    category=DeprecationWarning,
)



from typing import Dict, Any, Optional, List, Union
import numpy as np
from betterproto.lib.google.protobuf import Empty
from dependency_injector.wiring import inject, Provide

from qm.api.models.capabilities import ServerCapabilities
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.exceptions import ConfigValidationException
from qm.grpc import qua_config, octave_models
from qm.program._validate_config_schema import (
    validate_output_smearing,
    validate_oscillator,
    validate_output_tof,
    validate_arbitrary_waveform,
    validate_used_inputs,
    validate_timetagging_parameters,
)


class OctaveConnectionAmbiguity(Exception):
    def __str__(self):
        return (
            "It is not allowed to override the default connection of the Octave. You should either state the "
            "default connection to Octave in the controller level, or set each port separately in the port level."
        )


def octave_connection_to_pb(
    connection_type: str,
    *,
    device_name: Optional[str] = None,
    port_num: Optional[int] = None,
    port_data: Optional[dict] = None,
) -> Optional[Union[octave_models.OctaveIfInputPort, octave_models.OctaveIfOutputPort]]:
    if connection_type == "input":
        return_class, port_class = (
            octave_models.OctaveIfInputPort,
            octave_models.OctaveIfInputPortName,
        )
    elif connection_type == "output":
        return_class, port_class = (
            octave_models.OctaveIfOutputPort,
            octave_models.OctaveIfOutputPortName,
        )
    elif connection_type == "digital_input":
        return_class, port_class = (
            octave_models.OctaveDigitalInputPort,
            octave_models.OctaveDigitalInputPortName,
        )
    elif connection_type == "digital_output":
        return_class, port_class = (
            octave_models.OctaveDigitalOutputPort,
            octave_models.OctaveDigitalOutputPortName,
        )
    else:
        raise Exception("Unknown connection type")

    if device_name is not None:
        if isinstance(port_data, dict) and port_data.get("connectivity") is not None:
            raise OctaveConnectionAmbiguity
        port = port_class(port_num - 1)
        # The enum is 0-based and the port enumeration is 1-based, so we need to align them.
        return return_class(device_name=device_name, port_name=port)

    connectivity = port_data.get("connectivity")
    if connectivity is None:
        return None

    device_name, port_name = connectivity
    port = port_class[port_name]
    return return_class(device_name=device_name, port_name=port)


def analog_input_port_to_pb(
    data: Dict[str, Any],
    octave_output: Optional[octave_models.OctaveIfOutputPort],
) -> qua_config.QuaConfigAnalogInputPortDec:
    analog_input = qua_config.QuaConfigAnalogInputPortDec(
        offset=data.get("offset", 0.0),
        shareable=bool(data.get("shareable")),
        gain_db=int(data.get("gain_db", 0)),
    )

    if octave_output is not None:
        analog_input.octave_connectivity = octave_output

    return analog_input


def analog_output_port_to_pb(
    data: Dict[str, Any], octave_input: Optional[octave_models.OctaveIfInputPort]
) -> qua_config.QuaConfigAnalogOutputPortDec:
    analog_output = qua_config.QuaConfigAnalogOutputPortDec(shareable=bool(data.get("shareable")))

    if octave_input is not None:
        analog_output.octave_connectivity = octave_input

    if "offset" in data:
        analog_output.offset = data["offset"]

    if "delay" in data:
        delay = data.get("delay", 0)
        if delay < 0:
            raise ConfigValidationException(f"analog output delay cannot be a negative value, given value: {delay}")
        analog_output.delay = delay

    if "filter" in data:
        analog_output.filter = qua_config.QuaConfigAnalogOutputPortFilter(
            feedforward=data["filter"]["feedforward"],
            feedback=data["filter"]["feedback"],
        )

    if "crosstalk" in data:
        for k, v in data["crosstalk"].items():
            analog_output.crosstalk[int(k)] = v

    return analog_output


def digital_output_port_to_pb(data: Dict[str, Any]) -> qua_config.QuaConfigDigitalOutputPortDec:
    digital_output = qua_config.QuaConfigDigitalOutputPortDec(
        shareable=bool(data.get("shareable")),
        inverted=bool(data.get("inverted", False)),
    )
    # TODO - once we decide how to add digital connectivity this can be added back
    # if "connectivity" in data:
    #     digital_output.octave_connectivity = octave_connection_to_pb("digital_input", port_data=data)
    return digital_output


def digital_input_port_to_pb(data: Dict):
    digital_input = qua_config.QuaConfigDigitalInputPortDec(shareable=bool(data.get("shareable")))

    if "window" in data:
        digital_input.window = data["window"]

    if "threshold" in data:
        digital_input.threshold = data["threshold"]

    if "polarity" in data:
        if data["polarity"].upper() == "RISING":
            digital_input.polarity = qua_config.QuaConfigDigitalInputPortDecPolarity.RISING
        elif data["polarity"].upper() == "FALLING":
            digital_input.polarity = qua_config.QuaConfigDigitalInputPortDecPolarity.FALLING

    if "deadtime" in data:
        digital_input.deadtime = int(data["deadtime"])

    # TODO - once we decide how to add digital connectivity this can be added back
    # if "connectivity" in data:
    #     digital_input.octave_connectivity = octave_connection_to_pb("digital_output", port_data=data)

    return digital_input


def controller_to_pb(data: Dict[str, Any]) -> qua_config.QuaConfigControllerDec:
    cont = qua_config.QuaConfigControllerDec(type=data.get("type", "opx1"))

    if "type" in data:
        cont.type = data["type"]

    connected_octave_name = data.get("connectivity")
    if "analog_outputs" in data:
        for _k, _v in data["analog_outputs"].items():
            int_k = int(_k)
            octave_input = octave_connection_to_pb(
                "input", device_name=connected_octave_name, port_num=int_k, port_data=_v
            )
            cont.analog_outputs[int_k] = analog_output_port_to_pb(_v, octave_input)

    if "analog_inputs" in data:
        for _k, _v in data["analog_inputs"].items():
            int_k = int(_k)
            octave_input = octave_connection_to_pb(
                "output",
                device_name=connected_octave_name,
                port_num=int_k,
                port_data=_v,
            )
            cont.analog_inputs[int(_k)] = analog_input_port_to_pb(_v, octave_input)

    if "digital_outputs" in data:
        for _k, _v in data["digital_outputs"].items():
            cont.digital_outputs[int(_k)] = digital_output_port_to_pb(_v)

    if "digital_inputs" in data:
        for _k, _v in data["digital_inputs"].items():
            cont.digital_inputs[int(_k)] = digital_input_port_to_pb(_v)

    return cont


def get_octave_loopbacks(data: List) -> List[octave_models.OctaveLoopback]:
    loopbacks = [
        octave_models.OctaveLoopback(
            lo_source_input=octave_models.OctaveLoSourceInput[loopback[1]],
            lo_source_generator=octave_models.OctaveSynthesizerPort(
                device_name=loopback[0][0],
                port_name=octave_models.OctaveSynthesizerOutputName[loopback[0][1]],
            ),
        )
        for loopback in data
    ]
    return loopbacks


def octave_to_pb(data: Dict[str, Any]) -> octave_models.OctaveConfig:
    loopbacks = get_octave_loopbacks(data["loopbacks"])
    return octave_models.OctaveConfig(loopbacks)


@inject
def mixer_ref_to_pb(
    name: str,
    lo_frequency: int,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> qua_config.QuaConfigMixerRef:
    if capabilities.supports_double_frequency:
        frequency = {"lo_frequency": int(lo_frequency)}
    else:
        frequency = {"lo_frequency_double": float(lo_frequency)}

    return qua_config.QuaConfigMixerRef(mixer=name, **frequency)


@inject
def oscillator_to_pb(
    data, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
) -> qua_config.QuaConfigOscillator:
    oscillator = qua_config.QuaConfigOscillator()
    if "intermediate_frequency" in data:
        if capabilities.supports_double_frequency:
            oscillator.intermediate_frequency_double = float(data["intermediate_frequency"])
        else:
            oscillator.intermediate_frequency = int(data["intermediate_frequency"])

    if "mixer" in data:
        oscillator.mixer = qua_config.QuaConfigMixerRef(mixer=data["mixer"])
        if capabilities.supports_double_frequency:
            oscillator.mixer.lo_frequency_double = float(data.get("lo_frequency", 0.0))
        else:
            oscillator.mixer.lo_frequency = int(data.get("lo_frequency", 0))
    return oscillator


@inject
def create_correction_entry(
    mixer_data,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> qua_config.QuaConfigCorrectionEntry:
    correction = qua_config.QuaConfigCorrectionEntry(
        frequency_negative=mixer_data["intermediate_frequency"] < 0,
        correction=qua_config.QuaConfigMatrix(
            v00=mixer_data["correction"][0],
            v01=mixer_data["correction"][1],
            v10=mixer_data["correction"][2],
            v11=mixer_data["correction"][3],
        ),
    )
    if capabilities.supports_double_frequency:
        correction.frequency_double = abs(float(mixer_data["intermediate_frequency"]))
        correction.lo_frequency_double = float(mixer_data["lo_frequency"])
    else:
        correction.frequency = abs(int(mixer_data["intermediate_frequency"]))
        correction.lo_frequency = int(mixer_data["lo_frequency"])
    return correction


def mixer_to_pb(data) -> qua_config.QuaConfigMixerDec:
    return qua_config.QuaConfigMixerDec(correction=[create_correction_entry(mixer) for mixer in data])


def element_thread_to_pb(name: str) -> qua_config.QuaConfigElementThread:
    return qua_config.QuaConfigElementThread(thread_name=name)


def dac_port_ref_to_pb(controller: str, number: int) -> qua_config.QuaConfigDacPortReference:
    return qua_config.QuaConfigDacPortReference(controller=controller, number=number)


def single_input_to_pb(controller: str, number: int) -> qua_config.QuaConfigSingleInput:
    return qua_config.QuaConfigSingleInput(port=dac_port_ref_to_pb(controller, number))


def adc_port_ref_to_pb(controller: str, number: int) -> qua_config.QuaConfigAdcPortReference:
    return qua_config.QuaConfigAdcPortReference(controller=controller, number=number)


def port_ref_to_pb(controller: str, number: int) -> qua_config.QuaConfigPortReference:
    return qua_config.QuaConfigPortReference(controller=controller, number=number)


def digital_input_port_ref_to_pb(data) -> qua_config.QuaConfigDigitalInputPortReference:
    digital_input = qua_config.QuaConfigDigitalInputPortReference(
        delay=int(data["delay"]),
        buffer=int(data["buffer"]),
    )
    if "port" in data:
        digital_input.port = port_ref_to_pb(data["port"][0], data["port"][1])

    return digital_input


def digital_output_port_ref_to_pb(
    data,
) -> qua_config.QuaConfigDigitalOutputPortReference:
    return qua_config.QuaConfigDigitalOutputPortReference(port=port_ref_to_pb(data[0], data[1]))


def octave_params_to_pb(octave_params: dict) -> qua_config.QuaConfigUpConverted:
    pb_instance = qua_config.QuaConfigUpConverted(
        rf_output_port=octave_models.OctaveOutputPort(
            octave_params["rf_output_port"][0],
            octave_models.OctaveRfOutputPortName[octave_params["rf_output_port"][1]],
        ),
        lo_source=octave_models.OctaveLoSourceInput[octave_params["lo_source"]],
        output_switch_state=qua_config.QuaConfigOutputSwitchState[octave_params.get("output_switch_state", "unset")],
        output_gain=octave_params.get("output_gain", 0),
        downconversion_lo_frequency=octave_params.get("downconversion_lo_frequency", 0.0),
    )
    if "downconversion_lo_source" in octave_params:
        pb_instance.downconversion_lo_source = octave_models.OctaveLoSourceInput[
            octave_params["downconversion_lo_source"]
        ]
    if "rf_input_port" in octave_params:
        pb_instance.rf_input_port = octave_models.OctaveRfInputPort(
            octave_params["rf_input_port"][0],
            octave_models.OctaveRfInputPortName[octave_params["rf_input_port"][1]],
        )
    return pb_instance


@inject
def element_to_pb(
    element_name,
    data,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> qua_config.QuaConfigElementDec():
    validate_oscillator(data)
    validate_output_smearing(data)
    validate_output_tof(data)
    validate_timetagging_parameters(data)
    validate_used_inputs(data)

    element = qua_config.QuaConfigElementDec()

    if "time_of_flight" in data:
        element.time_of_flight = int(data["time_of_flight"])

    if "smearing" in data:
        element.smearing = int(data["smearing"])

    if "intermediate_frequency" in data:
        if capabilities.supports_double_frequency:
            element.intermediate_frequency_double = abs(float(data["intermediate_frequency"]))
            element.intermediate_frequency_oscillator_double = float(data["intermediate_frequency"])
        else:
            element.intermediate_frequency = abs(int(data["intermediate_frequency"]))
            element.intermediate_frequency_oscillator = int(data["intermediate_frequency"])

        element.intermediate_frequency_negative = data["intermediate_frequency"] < 0

    if "thread" in data:
        element.thread = element_thread_to_pb(data["thread"])

    if "outputs" in data:
        for k, v in data["outputs"].items():
            element.outputs[k] = adc_port_ref_to_pb(v[0], v[1])

    if "digitalInputs" in data:
        for k, v in data["digitalInputs"].items():
            element.digital_inputs[k] = digital_input_port_ref_to_pb(v)

    if "digitalOutputs" in data:
        for k, v in data["digitalOutputs"].items():
            element.digital_outputs[k] = digital_output_port_ref_to_pb(v)

    if "operations" in data:
        for k, v in data["operations"].items():
            element.operations[k] = v

    if "singleInput" in data:
        (cont, port_id) = data["singleInput"]["port"]
        element.single_input = single_input_to_pb(cont, port_id)

    if "mixInputs" in data:
        mix_inputs = data["mixInputs"]
        (cont_I, port_id_I) = mix_inputs["I"]
        (cont_Q, port_id_Q) = mix_inputs["Q"]
        element.mix_inputs = qua_config.QuaConfigMixInputs(
            i=dac_port_ref_to_pb(cont_I, port_id_I),
            q=dac_port_ref_to_pb(cont_Q, port_id_Q),
            mixer=mix_inputs.get("mixer", ""),
        )

        lo_frequency = mix_inputs.get("lo_frequency", 0)
        if capabilities.supports_double_frequency:
            element.mix_inputs.lo_frequency_double = float(lo_frequency)
        else:
            element.mix_inputs.lo_frequency = int(lo_frequency)

        if "octave_params" in mix_inputs:
            element.mix_inputs.octave_params = octave_params_to_pb(mix_inputs["octave_params"])

    if "singleInputCollection" in data:
        element.single_input_collection = qua_config.QuaConfigSingleInputCollection(
            inputs={k: dac_port_ref_to_pb(v[0], v[1]) for k, v in data["singleInputCollection"]["inputs"].items()}
        )

    if "multipleInputs" in data:
        element.multiple_inputs = qua_config.QuaConfigMultipleInputs(
            inputs={k: dac_port_ref_to_pb(v[0], v[1]) for k, v in data["multipleInputs"]["inputs"].items()}
        )

    if "oscillator" in data:
        element.named_oscillator = data["oscillator"]
    elif "intermediate_frequency" not in data:
        element.no_oscillator = Empty()

    if "sticky" in data:
        if capabilities.supports_sticky_elements:
            element.sticky = qua_config.QuaConfigSticky(
                analog=data["sticky"].get("analog", True),
                digital=data["sticky"].get("digital", False),
                duration=data["sticky"].get("duration", 1),
            )
        else:
            if "digital" in data["sticky"] and data["sticky"]["digital"]:
                raise ConfigValidationException(
                    f"Server does not support digital sticky used in element " f"'{element_name}'"
                )
            element.hold_offset = qua_config.QuaConfigHoldOffset(duration=data["sticky"].get("duration", 1))

    elif "hold_offset" in data:
        if capabilities.supports_sticky_elements:
            element.sticky = qua_config.QuaConfigSticky(
                analog=True,
                digital=False,
                duration=data["hold_offset"].get("duration", 1),
            )
        else:
            element.hold_offset = qua_config.QuaConfigHoldOffset(duration=data["hold_offset"]["duration"])

    if "outputPulseParameters" in data:
        pulse_parameters = data["outputPulseParameters"]
        output_pulse_parameters = qua_config.QuaConfigOutputPulseParameters(
            signal_threshold=pulse_parameters["signalThreshold"],
        )

        signal_polarity = pulse_parameters["signalPolarity"].upper()
        if signal_polarity == "ABOVE" or signal_polarity == "ASCENDING":
            output_pulse_parameters.signal_polarity = qua_config.QuaConfigOutputPulseParametersPolarity.ASCENDING
        elif signal_polarity == "BELOW" or signal_polarity == "DESCENDING":
            output_pulse_parameters.signal_polarity = qua_config.QuaConfigOutputPulseParametersPolarity.DESCENDING

        if "derivativeThreshold" in pulse_parameters:
            output_pulse_parameters.derivative_threshold = pulse_parameters["derivativeThreshold"]
            polarity = pulse_parameters["derivativePolarity"].upper()
            if polarity == "ABOVE" or polarity == "ASCENDING":
                output_pulse_parameters.derivative_polarity = (
                    qua_config.QuaConfigOutputPulseParametersPolarity.ASCENDING
                )
            elif polarity == "BELOW" or polarity == "DESCENDING":
                output_pulse_parameters.derivative_polarity = (
                    qua_config.QuaConfigOutputPulseParametersPolarity.DESCENDING
                )

        element.output_pulse_parameters = output_pulse_parameters
    return element


def waveform_to_pb(data) -> qua_config.QuaConfigWaveformDec:
    wf = qua_config.QuaConfigWaveformDec()
    if data["type"] == "constant":
        wf.constant = qua_config.QuaConfigConstantWaveformDec(sample=data["sample"])
    elif data["type"] == "arbitrary":
        is_overridable = data.get("is_overridable", False)
        has_max_allowed_error = "max_allowed_error" in data
        has_sampling_rate = "sampling_rate" in data
        validate_arbitrary_waveform(is_overridable, has_max_allowed_error, has_sampling_rate)

        wf.arbitrary = qua_config.QuaConfigArbitraryWaveformDec(samples=data["samples"], is_overridable=is_overridable)

        if has_max_allowed_error:
            wf.arbitrary.max_allowed_error = data["max_allowed_error"]
        elif has_sampling_rate:
            wf.arbitrary.sampling_rate = data["sampling_rate"]
        elif not is_overridable:
            wf.arbitrary.max_allowed_error = 1e-4
    return wf


def digital_waveform_to_pb(data) -> qua_config.QuaConfigDigitalWaveformDec:
    return qua_config.QuaConfigDigitalWaveformDec(
        samples=[qua_config.QuaConfigDigitalWaveformSample(value=bool(s[0]), length=s[1]) for s in data["samples"]]
    )


def pulse_to_pb(data) -> qua_config.QuaConfigPulseDec:
    pulse = qua_config.QuaConfigPulseDec()

    if "length" in data:
        pulse.length = int(data["length"])

    if "operation" in data:
        if data["operation"] == "control":
            pulse.operation = qua_config.QuaConfigPulseDecOperation.CONTROL
        else:
            pulse.operation = qua_config.QuaConfigPulseDecOperation.MEASUREMENT

    if "digital_marker" in data:
        pulse.digital_marker = data["digital_marker"]

    if "integration_weights" in data:
        for k, v in data["integration_weights"].items():
            pulse.integration_weights[k] = v

    if "waveforms" in data:
        if "single" in data["waveforms"]:
            pulse.waveforms["single"] = data["waveforms"]["single"]

        elif "I" in data["waveforms"]:
            pulse.waveforms["I"] = data["waveforms"]["I"]
            pulse.waveforms["Q"] = data["waveforms"]["Q"]
    return pulse


def build_iw_sample(data) -> List[qua_config.QuaConfigIntegrationWeightSample]:
    if len(data) > 0 and not isinstance(data[0], tuple):
        data = np.round(2**-15 * np.round(np.array(data) / 2**-15), 20)
        new_data = []
        for s in data:
            if len(new_data) == 0 or new_data[-1][0] != s:
                new_data.append((s, 4))
            else:
                new_data[-1] = (new_data[-1][0], new_data[-1][1] + 4)
        data = new_data
    return [qua_config.QuaConfigIntegrationWeightSample(value=s[0], length=int(s[1])) for s in data]


def integration_weights_to_pb(data) -> qua_config.QuaConfigIntegrationWeightDec:
    iw = qua_config.QuaConfigIntegrationWeightDec(
        cosine=build_iw_sample(data["cosine"]), sine=build_iw_sample(data["sine"])
    )
    return iw


def load_config_pb(config) -> qua_config.QuaConfig:

    pb_config = qua_config.QuaConfig(v1_beta=qua_config.QuaConfigQuaConfigV1())

    def set_controllers():
        for k, v in config["controllers"].items():
            pb_config.v1_beta.controllers[k] = controller_to_pb(v)

    def set_octaves():
        for k, v in config.get("octaves", {}).items():
            pb_config.v1_beta.octaves[k] = octave_to_pb(v)

    def set_elements():
        for k, v in config["elements"].items():
            pb_config.v1_beta.elements[k] = element_to_pb(k, v)

    def set_pulses():
        for k, v in config["pulses"].items():
            pb_config.v1_beta.pulses[k] = pulse_to_pb(v)

    def set_waveforms():
        for k, v in config["waveforms"].items():
            pb_config.v1_beta.waveforms[k] = waveform_to_pb(v)

    def set_digital_waveforms():
        for k, v in config["digital_waveforms"].items():
            pb_config.v1_beta.digital_waveforms[k] = digital_waveform_to_pb(v)

    def set_integration_weights():
        for k, v in config["integration_weights"].items():
            pb_config.v1_beta.integration_weights[k] = integration_weights_to_pb(v)

    def set_mixers():
        for k, v in config["mixers"].items():
            pb_config.v1_beta.mixers[k] = mixer_to_pb(v)

    def set_oscillators():
        for k, v in config["oscillators"].items():
            pb_config.v1_beta.oscillators[k] = oscillator_to_pb(v)

    key_to_action = {
        "version": lambda: None,
        "controllers": set_controllers,
        "elements": set_elements,
        "pulses": set_pulses,
        "waveforms": set_waveforms,
        "digital_waveforms": set_digital_waveforms,
        "integration_weights": set_integration_weights,
        "mixers": set_mixers,
        "oscillators": set_oscillators,
        "octaves": set_octaves,
    }

    for key in config:
        key_to_action[key]()

    return pb_config



import abc
from dataclasses import dataclass
from typing import Union, TypeVar, Generic, NewType, List

from qm.grpc.frontend import (
    SimulationRequest,
    ExecutionRequestSimulateSimulationInterfaceNone,
)


class SimulationConfig(object):
    """Creates a configuration object to pass to
    [qm.QuantumMachinesManager.QuantumMachinesManager.simulate][]

    Args:
        duration (int): The duration to run the simulation for, in clock
            cycles
        include_analog_waveforms (bool): True to collect simulated
            analog waveform names
        include_digital_waveforms (bool): True to collect simulated
            digital waveform names
        simulation_interface (SimulatorInterface):
            Interface for to simulator. Currently supported interfaces
            - ``None`` - Zero inputs
            - [qm.simulate.loopback.LoopbackInterface][] - Loopback output to input
            - [qm.simulate.raw.RawInterface][] - Specify samples for inputs
        controller_connections (List[ControllerConnection]): A list of
            connections between the controllers in the config
        extraProcessingTimeoutInMs (int): timeout in ms to wait for
            stream processing to finish. Default is -1, which is
            disables the timeout

    """

    duration = 0
    simulate_analog_outputs = False

    def __init__(
        self,
        duration=0,
        include_analog_waveforms=False,
        include_digital_waveforms=False,
        simulation_interface=None,
        controller_connections: List["ControllerConnection"] = None,
        extraProcessingTimeoutInMs=-1,
    ):
        if controller_connections is None:
            controller_connections = []
        super(SimulationConfig, self).__init__()
        self.duration = duration
        self.include_analog_waveforms = include_analog_waveforms is True
        self.include_digital_waveforms = include_digital_waveforms is True
        self.simulation_interface = simulation_interface
        self.controller_connections = controller_connections
        self.extraProcessingTimeoutInMs = extraProcessingTimeoutInMs

    def update_simulate_request(self, request: SimulationRequest):
        if self.simulation_interface is None:
            request.simulate.simulation_interface.none = ExecutionRequestSimulateSimulationInterfaceNone()
        else:
            request = self.simulation_interface.update_simulate_request(request)
        return request


class SimulatorInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def update_simulate_request(self, request: SimulationRequest) -> SimulationRequest:
        pass


@dataclass
class InterOpxAddress:
    """Args:
    controller (str): The name of the controller
    is_left_connection (bool)
    """

    controller: str
    is_left_connection: bool


@dataclass
class InterOpxChannel:
    """Args:
    controller (str): The name of the controller
    channel_number (int)
    """

    controller: str
    channel_number: int


InterOpxPairing = NewType("InterOpxPairing", Union[InterOpxAddress, InterOpxChannel])
T = TypeVar("T", InterOpxPairing, InterOpxPairing)


class ControllerConnection(Generic[T]):
    """"""

    def __init__(self, source: T, target: T):
        self.source = source
        self.target = target



from qm.jobs.pending_job import QmPendingJob  # noqa
import warnings

warnings.warn(
    "'qm.QmPendingJob.QmPendingJob' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.QmPendingJob' instead",
    category=DeprecationWarning,
)



import logging
from typing import Optional, Dict

from grpclib import GRPCError

from qm.api.models.debug_data import DebugData
from qm.api.models.server_details import (
    ConnectionDetails,
    ServerDetails,
    MAX_MESSAGE_SIZE,
    BASE_TIMEOUT,
)
from qm.exceptions import QmServerDetectionError, QMTimeoutError
from qm.api.frontend_api import FrontendApi
from qm.api.info_service_api import InfoServiceApi

logger = logging.getLogger(__name__)

DEFAULT_PORTS = (80, 9510)


def _fill_headers(headers: Dict[str, str], user_token: Optional[str]) -> Dict[str, str]:
    headers.update({"x-grpc-service": "gateway"})
    if user_token:
        headers["authorization"] = f"Bearer {user_token}"
    return headers


def _create_server_info(
    user_token: Optional[str],
    credentials: Optional[str],
    host: str,
    port: int,
    add_debug_data: bool,
    headers: Dict[str, str],
    timeout: Optional[float] = None,
    max_message_size: Optional[int] = None,
) -> ServerDetails:
    debug_data = DebugData()

    connection_details = ConnectionDetails(
        host=host,
        port=port,
        user_token=user_token,
        credentials=credentials,
        max_message_size=max_message_size if max_message_size else MAX_MESSAGE_SIZE,
        headers=_fill_headers(headers, user_token),
        timeout=timeout if timeout else BASE_TIMEOUT,
        debug_data=debug_data if add_debug_data else None,
    )

    frontend = FrontendApi(connection_details)
    info_service = InfoServiceApi(connection_details)

    try:
        qop_version = frontend.get_version()
    except (GRPCError, OSError, QMTimeoutError):
        qop_version = None

    try:
        info = info_service.get_info()
    except (GRPCError, OSError, QMTimeoutError):
        info = None

    details = ServerDetails(
        port=port,
        host=host,
        qop_version=qop_version,
        qua_implementation=info,
        connection_details=connection_details,
    )

    return details


def detect_server(
    user_token: Optional[str],
    credentials: Optional[str],
    host: str,
    port_from_user_config: int,
    user_provided_port: Optional[int],
    add_debug_data: bool,
    timeout: Optional[float] = None,
    max_message_size: Optional[int] = None,
    extra_headers: Optional[Dict[str, str]] = None,
) -> ServerDetails:
    if user_provided_port is None:
        possible_ports_to_try = [port_from_user_config, *DEFAULT_PORTS]
        ports_to_try = []
        for port in possible_ports_to_try:
            if port is not None:
                if int(port) not in ports_to_try:
                    ports_to_try.append(int(port))
    else:
        ports_to_try = [user_provided_port]

    if extra_headers is None:
        extra_headers = {}

    for port in ports_to_try:
        detected = _create_server_info(
            user_token,
            credentials,
            host,
            port,
            add_debug_data,
            extra_headers,
            timeout,
            max_message_size,
        )
        if detected.qop_version is not None:
            return detected

    targets = ",".join([f"{host}:{port}" for port in ports_to_try])
    message = f"Failed to detect to QuantumMachines server. Tried connecting to {targets}."
    logger.error(message)
    raise QmServerDetectionError(message)



import warnings
from typing import Dict, Tuple, Union, List, Optional, ContextManager

from octave_sdk.octave import ClockInfo
from qm.elements.element_with_octave import ElementWithOctave

from qm.elements_db import ElementsDB
from qm.octave.calibration_db import CalibrationResult
from octave_sdk import (
    OctaveLOSource,
    RFOutputMode,
    RFInputLOSource,
    IFMode,
    ClockType,
    ClockFrequency,
)
from qm.octave.octave_manager import OctaveManager, ClockMode


class ElementHasNoOctaveError(Exception):
    pass


class QmOctave:
    def __init__(self, elements_db: ElementsDB, octave_manager) -> None:
        super().__init__()
        self._elements_db = elements_db
        self._octave_manager: OctaveManager = octave_manager

    def start_batch_mode(self):
        self._octave_manager.start_batch_mode()

    def end_batch_mode(self):
        self._octave_manager.end_batch_mode()

    def batch_mode(self) -> ContextManager:
        return self._octave_manager.batch_mode()

    def set_qua_element_octave_rf_in_port(self, element, octave_name, rf_input_index):
        """Sets the octave downconversion port for the element.

        Args:
            element (str): The name of the element
            octave_name (str): The octave name
            rf_input_index (RFInputRFSource): input index
        """
        inst = self._elements_db[element]
        client = self._octave_manager._octave_clients[octave_name]
        inst.set_downconversion_port(client, rf_input_index)

    def load_lo_frequency_from_config(self, elements: Union[List, str]):
        """Loads into the octave synthesizers the LO frequencies specified for elements
        in the element list

        Args:
            elements: A list of elements to load LO frequencies from
        """
        warnings.warn(
            "lo_frequency is now set directly from config when a QuantumMachine is created, no need to do it directly. "
            "If you want, you can run over the elements and do set_frequency.",
            category=DeprecationWarning,
        )

    def set_lo_frequency(self, element: str, lo_frequency: float, set_source: bool = True):
        """Sets the LO frequency of the synthesizer associated to element. Will not change the synthesizer if set_source = False

        Args:
            element (str): The name of the element
            lo_frequency (float): The LO frequency
            set_source (Boolean): Set the synthesizer (True) or just
                update the client (False)
        """
        inst = self._elements_db[element]
        inst.set_lo_frequency(lo_frequency, set_source)

    def update_external_lo_frequency(
        self,
        element: str,
        lo_frequency: float,
    ):
        """Updates the client on the external LO frequency (provided by the user)
        associated with element

        Args:
            element (str): The name of the element
            lo_frequency (float): The LO frequency
        """
        self._elements_db[element].lo_frequency = lo_frequency
        self._elements_db[element].set_lo_frequency(lo_frequency)

    def set_lo_source(self, element: str, lo_port: OctaveLOSource):
        """Associate  the given LO source with the given element. Always be sure the given LO source is internally connected to the element

        Args:
            element (str): The name of the element
            lo_port (OctaveLOSource): One of the allowed sources
                according the internal connectivity
        """
        inst = self._elements_db[element]
        if not isinstance(inst, ElementWithOctave):
            raise ElementHasNoOctaveError(f"Element {element} has no octave connected to it.")
        inst.set_lo_source(lo_port)

    def set_rf_output_mode(self, element: str, switch_mode: RFOutputMode):
        """Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'on', 'off', 'trig_normal' or 'trig_inverse'
        When in 'trig_normal' mode a high trigger will turn the switch on and a low trigger will turn it off
        When in 'trig_inverse' mode a high trigger will turn the switch off and a low trigger will turn it on
        When in 'on' the switch will be permanently on. When in 'off' mode the switch will be permanently off.

        Args:
            element (str): The name of the element
            switch_mode (RFOutputMode): switch mode according to the
                allowed states
        """
        inst = self._elements_db[element]
        inst.set_rf_output_mode(switch_mode)

    def set_rf_output_gain(self, element: str, gain_in_db: float):
        """Sets the RF output gain for the up-converter associated with the element.
        RF_gain is in steps of 0.5dB from -20 to +24 and is referring
        to the maximum OPX output power of 4dBm (=0.5V pk-pk) So for a value of -24
        for example, an IF signal coming from the OPX at max power (4dBm) will be
        upconverted and come out of Octave at -20dBm

        Args:
            element (str): The name of the element
            gain_in_db (float): The RF output gain in dB
        """
        # a. use value custom set in qm.octave.update_external
        # b. use value from config
        inst = self._elements_db[element]
        inst.set_rf_output_gain(gain_in_db)

    def set_downconversion(
        self,
        element: str,
        lo_source: Optional[RFInputLOSource] = None,
        lo_frequency: Optional[float] = None,
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning: bool = False,
    ):
        """Sets the LO source and frequency for the downconverters.
        The LO source will be the one associated with the element's upconversion.
        If only the element is given, the LO source for downconversion will be the upconversion LO of the element.

        Args:
            element (str): The name of the element
            lo_source (RFInputLOSource): allowed LO source
            lo_frequency (float): The LO frequency
            disable_warning (Boolean): Disable warnings about non-
                matching LO sources and elements if True
        """
        inst = self._elements_db[element]
        inst.set_downconversion(lo_source, lo_frequency, if_mode_i, if_mode_q)

    def calibrate_element(
        self,
        element: str,
        lo_if_frequencies_tuple_list: Optional[List[Tuple]] = None,
        save_to_db=True,
        close_open_quantum_machines=True,
        **kwargs,
    ) -> Dict[Tuple[int, int], CalibrationResult]:
        """Calibrate the mixer associated with an element for the given LO & IF frequencies.

        Args:
            close_open_quantum_machines (bool): If true (default) all
                running QMs will be closed for the calibration.
                Otherwise, calibration might fail if there are not
                enough resources for the calibration
            element (str): The name of the element for calibration
            lo_if_frequencies_tuple_list (list): a list of tuples that
                consists of all the (LO,IF) pairs for calibration
            save_to_db (boolean): If true (default), The calibration
                parameters will be saved to the calibration database
        """
        inst = self._elements_db[element]
        octave_port = (inst._client._client._octave_name, inst._octave_if_input_port_number)

        # get IF and LO frequencies from element
        if lo_if_frequencies_tuple_list is None:
            if_frequency = inst.intermediate_frequency
            lo_frequency = inst.lo_frequency
            lo_if_frequencies_tuple_list = [(if_frequency, lo_frequency)]

        return self._octave_manager.calibrate(
            octave_port,
            lo_if_frequencies_tuple_list,
            save_to_db,
            close_open_quantum_machines,
            **kwargs,
        )

    def set_clock(
        self,
        octave_name: str,
        clock_type: Optional[ClockType] = None,
        frequency: Optional[ClockFrequency] = None,
        clock_mode: Optional[ClockMode] = None,
    ):
        """This function will set the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        Args:
            octave_name (str): The octave name to set clock for
            clock_type (ClockType): clock type according to ClockType
            frequency (ClockFrequency): Clock frequency according to ClockFrequency
            clock_mode (ClockMode):
        """
        # TODO: get name from the QMOctave instance
        self._octave_manager.set_clock(octave_name, clock_type, frequency, clock_mode)

    def get_clock(self, octave_name: str) -> ClockInfo:
        """Gets the clock info for a given octave name

        Args:
            octave_name (str): The octave name to get clock for
        :returns ClockInfo: Info about the clock as an object
        """
        return self._octave_manager.get_clock(octave_name)



from dataclasses import dataclass, field
from typing import Optional, List

from qm.grpc.qua import QuaProgramCompilerOptions


@dataclass
class CompilerOptionArguments:
    strict: Optional[bool] = field(default=None)

    flags: List[str] = field(default_factory=list)


def _get_request_compiler_options(
    compiler_options: CompilerOptionArguments,
) -> QuaProgramCompilerOptions:
    flags = compiler_options.flags
    if compiler_options.strict:
        flags.append("strict")

    return QuaProgramCompilerOptions(flags=compiler_options.flags)



from qm.jobs.qm_job import QmJob  # noqa
import warnings

warnings.warn(
    "'qm.QmJob.QmJob' is moved as of 1.1.0 and will be removed in 1.2.0. " "use 'qm.QmJob' instead",
    category=DeprecationWarning,
)



import logging

from qm.version import __version__  # noqa

from qm.QuantumMachinesManager import QuantumMachinesManager  # noqa
from qm.QuantumMachine import QuantumMachine  # noqa

from qm.jobs.job_queue import QmQueue  # noqa
from qm.jobs.pending_job import QmPendingJob  # noqa
from qm.jobs.qm_job import QmJob  # noqa

from qm.results import (  # noqa
    StreamingResultFetcher,
    SingleStreamingResultFetcher,
    MultipleStreamingResultFetcher,
)

from qm.program import Program, _ResultAnalysis, _Program  # noqa

from qm.api.models.compiler import CompilerOptionArguments  # noqa
from qm.serialization.generate_qua_script import generate_qua_script  # noqa
from qm.logging_utils import config_loggers
from qm.user_config import UserConfig  # noqa

from qm.simulate import (  # noqa
    SimulationConfig,
    InterOpxAddress,
    InterOpxChannel,
    ControllerConnection,
    InterOpxPairing,
    LoopbackInterface,
)

from qm.type_hinting import DictQuaConfig

__all__ = [
    "QuantumMachinesManager",
    "QuantumMachine",
    "QmQueue",
    "QmPendingJob",
    "QmJob",
    "StreamingResultFetcher",
    "SingleStreamingResultFetcher",
    "MultipleStreamingResultFetcher",
    "Program",
    "_ResultAnalysis",
    "CompilerOptionArguments",
    "generate_qua_script",
    "config_loggers",
    "UserConfig",
    "SimulationConfig",
    "InterOpxAddress",
    "InterOpxChannel",
    "ControllerConnection",
    "InterOpxPairing",
    "LoopbackInterface",
    "DictQuaConfig",
]

config = UserConfig.create_from_file()
config_loggers(config)


logger = logging.getLogger(__name__)
logger.info(f"Starting session: {config.SESSION_ID}")



import logging
from logging import config
from qm.user_config import UserConfig


def config_loggers(user_config: UserConfig):
    config.dictConfig(user_config.logging_config_dict)


def set_logging_level(level: int) -> None:
    """Sets the logging level of the qm-qua module ("qm")
    See `Messages control <https://qm-docs.qualang.io/guides/error#messages-control>`__ for more information.

    Args:
        level: A string of either: 'DEBUG', 'INFO', 'WARNING', or
            'ERROR'. Can also accept standard python logging levels.
    """
    package_logger = logging.getLogger("qm")
    try:
        package_logger.setLevel(level)
    except (ValueError, TypeError):
        package_logger.warning(f"Failed to set log level. level '{level}' is not recognized")



from qm.grpc.frontend import (
    SimulationRequest,
    ExecutionRequestSimulateSimulationInterfaceLoopback,
    ExecutionRequestSimulateSimulationInterfaceLoopbackConnections,
)
from qm.simulate.interface import SimulatorInterface


class LoopbackInterface(SimulatorInterface):
    """Creates a loopback interface for use in
    [qm.simulate.interface.SimulationConfig][].
    A loopback connects the output of the OPX into it's input. This can be defined
    directly using the ports or through the elements.

    Args:
        connections (list):
            List of tuples with loopback connections. Each tuple can be

                1. Physical connection between ports:

                    ``(fromController: str, fromPort: int, toController: str, toPort: int)``

                2. Virtual connection between elements:

                    ``(fromQE: str, toQE: str, toQEInput: int)``
        latency (int): The latency between the OPX outputs and its
            input.
        noisePower (float): How much noise to add to the input.

    Example:
        ```python
        job = qmm.simulate(config, prog, SimulationConfig(
                          duration=20000,
                          # loopback from output 1 to input 2 of controller 1:
                          simulation_interface=LoopbackInterface([("con1", 1, "con1", 2)])
        ```
    """

    def __init__(self, connections, latency=24, noisePower=0.0):
        if type(latency) is not int or latency < 0:
            raise Exception("latency must be a positive integer")

        self.latency = latency

        if (type(noisePower) is not float and type(noisePower) is not int) or noisePower < 0:
            raise Exception("noisePower must be a positive number")

        self.noisePower = noisePower

        if type(connections) is not list:
            raise Exception("connections argument must be of type list")

        self.connections = list()
        for connection in connections:
            if type(connection) is not tuple:
                raise Exception("each connection must be of type tuple")
            if len(connection) == 4:
                if (
                    type(connection[0]) is not str
                    or type(connection[1]) is not int
                    or type(connection[2]) is not str
                    or type(connection[3]) is not int
                ):
                    raise Exception("connection should be (fromController, fromPort, toController, toPort)")
                self.connections.append(connection)
            elif len(connection) == 3:
                if type(connection[0]) is not str or type(connection[1]) is not str or type(connection[2]) is not int:
                    raise Exception("connection should be (fromQE, toQE, toQEInput)")
                self.connections.append((connection[0], -1, connection[1], connection[2]))
            else:
                raise Exception("connection should be tuple of length 3 or 4")

    def update_simulate_request(self, request: SimulationRequest) -> SimulationRequest:
        request.simulate.simulation_interface.loopback = ExecutionRequestSimulateSimulationInterfaceLoopback(
            latency=self.latency, noise_power=self.noisePower
        )
        for connection in self.connections:
            con = ExecutionRequestSimulateSimulationInterfaceLoopbackConnections(
                from_controller=connection[0],
                from_port=connection[1],
                to_controller=connection[2],
                to_port=connection[3],
            )
            request.simulate.simulation_interface.loopback.connections.append(con)
        return request



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/octave.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class OctaveRfInputPortName(betterproto.Enum):
    RF_IN1 = 0
    RF_IN2 = 1


class OctaveIfOutputPortName(betterproto.Enum):
    IF1 = 0
    IF2 = 1


class OctaveIfInputPortName(betterproto.Enum):
    I1 = 0
    Q1 = 1
    I2 = 2
    Q2 = 3
    I3 = 4
    Q3 = 5
    I4 = 6
    Q4 = 7
    I5 = 8
    Q5 = 9


class OctaveSynthesizerOutputName(betterproto.Enum):
    Synth1 = 0
    Synth2 = 1
    Synth3 = 2
    Synth4 = 3
    Synth5 = 4


class OctaveRfOutputPortName(betterproto.Enum):
    RF1 = 0
    RF2 = 1
    RF3 = 2
    RF4 = 3
    RF5 = 4


class OctaveDigitalOutputPortName(betterproto.Enum):
    DigOut1 = 0
    DigOut2 = 1


class OctaveDigitalInputPortName(betterproto.Enum):
    DigIn1 = 0
    DigIn2 = 1
    DigIn3 = 2
    DigIn4 = 3
    DigIn5 = 4


class OctaveLoSourceInput(betterproto.Enum):
    Internal = 0
    Off = -1
    Dmd1LO = 1
    Dmd2LO = 2
    LO1 = 10
    LO2 = 11
    LO3 = 12
    LO4 = 13
    LO5 = 14
    IF1_LO_I = 22
    """RF_IN1 = 20;    RF_IN2 = 21;"""

    IF1_LO_Q = 23
    IF2_LO_I = 24
    IF2_LO_Q = 25


@dataclass(eq=False, repr=False)
class Octave(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class OctaveConfig(betterproto.Message):
    loopbacks: List["OctaveLoopback"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class OctaveLoopback(betterproto.Message):
    lo_source_input: "OctaveLoSourceInput" = betterproto.enum_field(1)
    lo_source_generator: "OctaveSynthesizerPort" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class OctaveIfOutputPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveIfOutputPortName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class OctaveIfInputPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveIfInputPortName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class OctaveRfInputPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveRfInputPortName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class OctaveSynthesizerPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveSynthesizerOutputName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class OctaveOutputPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveRfOutputPortName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class OctaveDigitalOutputPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveDigitalOutputPortName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class OctaveDigitalInputPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "OctaveDigitalInputPortName" = betterproto.enum_field(2)



import dataclasses
from typing import Optional, Dict

from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.info import QuaMachineInfo
from qm.api.models.debug_data import DebugData

MAX_MESSAGE_SIZE = 1024 * 1024 * 100  # 100 mb in bytes
BASE_TIMEOUT = 60


@dataclasses.dataclass
class ConnectionDetails:
    host: str
    port: int
    user_token: Optional[str]
    credentials: Optional[str]
    max_message_size: int = dataclasses.field(default=MAX_MESSAGE_SIZE)
    headers: Dict[str, str] = dataclasses.field(default_factory=dict)
    timeout: float = dataclasses.field(default=BASE_TIMEOUT)
    debug_data: Optional[DebugData] = dataclasses.field(default=None)


@dataclasses.dataclass
class ServerDetails:
    port: int
    host: str
    qop_version: Optional[str]
    connection_details: ConnectionDetails

    # does it implement the QUA service
    qua_implementation: Optional[QuaMachineInfo]
    capabilities: ServerCapabilities = dataclasses.field(default_factory=ServerCapabilities.build)

    def __post_init__(self) -> None:
        self.capabilities = ServerCapabilities.build(qua_implementation=self.qua_implementation)



import base64
import dataclasses
import datetime
import functools
import logging
import os.path

from typing import Dict, List, Optional, Any, Union, Tuple, Iterable
from abc import abstractmethod
from abc import ABCMeta
from dataclasses import dataclass

import numpy as np
import plotly.colors
import plotly.graph_objects as go


@dataclass(frozen=True)
class PlayedWaveform(metaclass=ABCMeta):
    waveform_name: str
    pulse_name: str
    length: int
    timestamp: int
    iq_info: Dict[str, Any]
    element: str
    output_ports: List[int]
    controller: str
    pulser: Dict[str, Any]

    @classmethod
    def from_job_dict(cls, dict_description: Dict, formated_attribute_dict: Dict):
        formated_attribute_dict.update(
            pulse_name=dict_description["pulseName"],
            waveform_name=dict_description["waveformName"],
            timestamp=int(dict_description["timestamp"]),
            length=int(dict_description["length"]),
            iq_info=dict_description["iqInfo"],
            element=dict_description["quantumElements"],
            output_ports=[int(p) for p in dict_description["outputPorts"]],
            pulser=dict_description["pulser"],
            controller=dict_description["pulser"]["controllerName"],
        )
        return cls(**formated_attribute_dict)

    @property
    def ports(self) -> List[int]:
        return self.output_ports

    @property
    def is_iq(self) -> bool:
        return self.iq_info["isPartOfIq"]

    @property
    def is_I_pulse(self) -> bool:
        return self.iq_info["isI"]

    @property
    def get_iq_association(self) -> str:
        if not self.is_iq:
            return ""
        return "I" if self.is_I_pulse else "Q"

    @property
    def ends_at(self) -> int:
        return self.timestamp + self.length

    @abstractmethod
    def to_string(self) -> str:
        return ""

    def __str__(self):
        return self.to_string()

    def to_dict(self):
        return dataclasses.asdict(self)

    def _common_attributes_to_printable_str_list(self) -> List[str]:
        waveform_type_string = "Type="
        if self.is_iq:
            waveform_type_string += f"IQ Type ({'I' if self.iq_info['isI']  else 'Q'})"
        else:
            waveform_type_string += "Single"
        return [
            f"Waveform Name={self.waveform_name}",
            f"Pulse name={self.pulse_name.removeprefix('OriginPulseName=')}",
            f"Start Time={self.timestamp} ns",
            f"Length={self.length} ns",
            f"Element={self.element}",
            f"Output Ports={self.output_ports}",
            waveform_type_string,
        ]


def format_float(f) -> str:
    return "{:.3f}".format(f)


def pretty_string_freq(f) -> str:
    if f < 1000:
        div, units = 1, "Hz"
    elif 1000 <= f < 1_000_000:
        div, units = 1000, "kHz"
    else:
        div, units = 10e6, "MHz"
    return f"{format_float(f/div).rstrip('0').rstrip('.')}{units}"


@dataclass(frozen=True)
class PlayedAnalogWaveform(PlayedWaveform):
    current_amp_elements: List[float]
    current_dc_offset_by_port: Dict[str, float]
    current_intermediate_frequency: float
    current_frame: List[float]
    current_correction_elements: List[float]
    chirp_info: Optional[Dict]
    current_phase: float

    @classmethod
    def from_job_dict(cls, dict_description: Dict):
        pulse_chirp_info = dict_description["chirpInfo"]
        is_pulse_have_chirp = len(pulse_chirp_info["units"]) > 0 or len(pulse_chirp_info["rate"]) > 0
        formated_attribute_list = dict(
            current_amp_elements=dict_description["currentGMatrixElements"],
            current_dc_offset_by_port=dict_description["currentDCOffsetByPort"],
            current_intermediate_frequency=dict_description["currentIntermediateFrequency"],
            current_frame=dict_description["currentFrame"],
            current_correction_elements=dict_description["currentCorrectionElements"],
            chirp_info=pulse_chirp_info if is_pulse_have_chirp else None,
            current_phase=dict_description.get("currentPhase", 0),
        )
        return super(cls, cls).from_job_dict(dict_description, formated_attribute_list)

    def _to_custom_string(self, show_chirp=True) -> str:
        _attributes = super()._common_attributes_to_printable_str_list()
        _attributes += (
            [
                f"{k}={v if self.is_iq else v[0]}"
                for k, v in [
                    (
                        "Amplitude Values",
                        [format_float(f) for f in self.current_amp_elements],
                    ),
                    (
                        "Frame Values",
                        [format_float(f) for f in self.current_frame],
                    ),
                    (
                        "Correction Values",
                        [format_float(f) for f in self.current_correction_elements],
                    ),
                ]
            ]
            + [
                f"Intermediate Frequency={pretty_string_freq(self.current_intermediate_frequency)}",
                f"Current DC Offset (By output ports)={ {k: format_float(v) for k, v in self.current_dc_offset_by_port.items()} }",
                f"Current Phase={format_float(self.current_phase)},",
            ]
            + ([] if (self.chirp_info is None or not show_chirp) else [f"chirp_info={self.chirp_info}"])
        )
        s = "AnalogWaveform(" + ("\n" + len("AnalogWaveform(") * " ").join(_attributes) + ")"
        return s

    def to_string(self) -> str:
        return self._to_custom_string()


@dataclass(frozen=True)
class PlayedDigitalWaveform(PlayedWaveform):
    @classmethod
    def from_job_dict(cls, dict_description: Dict):
        return super(cls, cls).from_job_dict(dict_description, {})

    def to_string(self) -> str:
        s = (
            "DigitalWaveform("
            + ("\n" + len("DigitalWaveform(") * " ").join(self._common_attributes_to_printable_str_list())
            + ")"
        )
        return s


@dataclass(frozen=True)
class AdcAcquisition:
    start_time: int
    end_time: int
    process: str
    pulser: Dict[str, dict]
    quantum_element: str
    adc_ports: List[int]
    controller: str

    @classmethod
    def from_job_dict(cls, dict_description: Dict):
        return cls(
            start_time=int(dict_description["startTime"]),
            end_time=int(dict_description["endTime"]),
            process=dict_description["process"],
            pulser=dict_description["pulser"],
            quantum_element=dict_description["quantumElement"],
            adc_ports=[int(p) + 1 for p in dict_description["adc"]],
            controller=dict_description["pulser"]["controllerName"],
        )

    @property
    def ports(self):
        return self.adc_ports

    def to_string(self) -> str:
        return (
            "AdcAcquisition("
            + ("\n" + len("AdcAcquisition(") * " ").join(
                [
                    f"start_time={self.start_time}",
                    f"end_time={self.end_time}",
                    f"process={self.process}",
                    f"element={self.quantum_element}",
                    f"input_ports={self.adc_ports}",
                ]
            )
            + ")"
        )

    def to_dict(self):
        return dataclasses.asdict(self)


class WaveformReport:
    def __init__(self, descriptor_dict: Optional[Dict] = None, job_id=-1):
        self.analog_waveforms: List[PlayedAnalogWaveform] = []
        self.digital_waveforms: List[PlayedDigitalWaveform] = []
        self.adc_acquisitions: List[AdcAcquisition] = []
        self.job_id = job_id
        if descriptor_dict is not None:
            for wf in descriptor_dict["analogWaveforms"]:
                self.analog_waveforms.append(PlayedAnalogWaveform.from_job_dict(wf))
            for wf in descriptor_dict["digitalWaveforms"]:
                self.digital_waveforms.append(PlayedDigitalWaveform.from_job_dict(wf))
            if "adcAcquisitions" in descriptor_dict:
                for acq in descriptor_dict["adcAcquisitions"]:
                    self.adc_acquisitions.append(AdcAcquisition.from_job_dict(acq))
        return

    @classmethod
    def from_dict(cls, d: Dict, job_id=-1) -> "WaveformReport":
        return cls(d, job_id)

    @property
    def waveforms(self) -> List[PlayedWaveform]:
        return self.analog_waveforms + self.digital_waveforms

    @property
    def controllers_in_use(self) -> List[str]:
        return list(set([c.controller for c in (self.waveforms + self.adc_acquisitions)]))

    @property
    def num_controllers_in_use(self) -> int:
        return len(self.controllers_in_use)

    def _sort_ports_dict(self, ports_in_use):
        for k, v in ports_in_use.items():
            ports_in_use[k] = sorted(v)
        return ports_in_use

    def _get_output_ports_in_use(self, src: List, on_controller):
        ports_in_use = {c: set() for c in self.controllers_in_use}
        for ap in src:
            ports_in_use[ap.controller].update(ap.ports)
        self._sort_ports_dict(ports_in_use)
        if on_controller is None:
            if self.num_controllers_in_use == 1:
                return ports_in_use[self.controllers_in_use[0]]
            else:
                return ports_in_use
        else:
            return ports_in_use[on_controller]

    def analog_output_ports_in_use(self, on_controller: Optional[str] = None) -> Union[List[int], Dict[str, List[int]]]:
        return self._get_output_ports_in_use(self.analog_waveforms, on_controller)

    def digital_output_ports_in_use(
        self, on_controller: Optional[str] = None
    ) -> Union[List[int], Dict[str, List[int]]]:
        return self._get_output_ports_in_use(self.digital_waveforms, on_controller)

    def adcs_ports_in_use(self, on_controller: Optional[str] = None) -> Union[List[int], Dict[str, List[int]]]:
        return self._get_output_ports_in_use(self.adc_acquisitions, on_controller)

    def to_string(self):
        return "\n".join([wf.to_string() for wf in (self.waveforms + self.adc_acquisitions)])

    def _transform_report_by_func(self, func) -> "WaveformReport":
        new_report = WaveformReport(job_id=self.job_id)
        new_report.analog_waveforms = list(filter(func, self.analog_waveforms))
        new_report.digital_waveforms = list(filter(func, self.digital_waveforms))
        new_report.adc_acquisitions = list(filter(func, self.adc_acquisitions))
        return new_report

    def report_by_controllers(self) -> Dict[str, "WaveformReport"]:
        by_controller_map = {k: None for k in self.controllers_in_use}

        def _filter_func(r, conn):
            return r.controller == conn

        for con_name in by_controller_map.keys():
            con_filter = functools.partial(_filter_func, conn=con_name)
            by_controller_map[con_name] = self._transform_report_by_func(con_filter)

        return by_controller_map

    def to_dict(self) -> Dict:
        return {
            "analog_waveforms": [awf.to_dict() for awf in self.analog_waveforms],
            "digital_waveforms": [dwf.to_dict() for dwf in self.digital_waveforms],
            "adc_acquisitions": [acq.to_dict() for acq in self.adc_acquisitions],
        }

    def get_report_by_output_ports(self, on_controller: Optional[str] = None):

        map_by_output_ports = {
            con_name: {
                "analog_out": {k: [] for k in self.analog_output_ports_in_use(on_controller=con_name)},
                "digital_out": {k: [] for k in self.digital_output_ports_in_use(on_controller=con_name)},
                "analog_in": {k: [] for k in self.adcs_ports_in_use(on_controller=con_name)},
            }
            for con_name in self.controllers_in_use
        }

        for wf in self.analog_waveforms:
            for p in wf.output_ports:
                map_by_output_ports[wf.controller]["analog_out"][p].append(wf)
        for wf in self.digital_waveforms:
            for p in wf.output_ports:
                map_by_output_ports[wf.controller]["digital_out"][p].append(wf)
        for adc in self.adc_acquisitions:
            for p in adc.adc_ports:
                map_by_output_ports[adc.controller]["analog_in"][p].append(adc)

        if on_controller is None:
            if self.num_controllers_in_use == 1:
                return map_by_output_ports[self.controllers_in_use[0]]
            else:
                return map_by_output_ports
        else:
            return map_by_output_ports[on_controller]

    def create_plot(
        self,
        samples: Optional[Iterable] = None,
        controllers: Optional[List[str]] = None,
        plot: bool = True,
        save_path: Optional[str] = None,
    ):

        if save_path is None:
            dirname, filename = "./", f"waveform_report_{self.job_id}"
        else:
            dirname, filename = os.path.split(save_path)
            if filename == "":
                filename = f"waveform_report_{self.job_id}"

        report_by_controllers = self.report_by_controllers()
        filter_func = (lambda item: item[0] in controllers) if controllers is not None else (lambda x: True)
        for con_name, report in filter(filter_func, report_by_controllers.items()):
            con_samples = None
            if samples is not None:
                con_samples = samples.__getattribute__(con_name)
            con_builder = _WaveformPlotBuilder(report, con_samples, self.job_id)
            con_builder.build()
            if save_path is not None:
                filename = f"waveform_report_{con_name}_{self.job_id}"
                con_builder.save(dirname, filename)
            if plot:
                con_builder.plot()
        return


class _WaveformPlotBuilder:
    def __init__(self, wf_report: WaveformReport, samples=None, job_id=-1):
        self._report = wf_report
        if wf_report.num_controllers_in_use > 1:
            raise RuntimeError(
                f"Plot Builder does not support plotting more than 1 controllers, yet. {os.linesep}"
                "Please provide a report containing a single controller."
            )
        self._samples = samples
        self._job_id = job_id
        self._figure: Union[go.Figure, None] = None
        self._already_registered_qe = set()
        self._colormap = {}
        self._max_parallel_traces_per_row = {}
        return

    @property
    def _num_rows(self) -> int:
        with_samples = self._samples is not None
        num_rows = (
            len(self._report.analog_output_ports_in_use()) + len(self._report.digital_output_ports_in_use())
        ) * (1 + with_samples)
        num_rows += len(self._report.adcs_ports_in_use())
        return num_rows

    @property
    def _num_output_rows(self) -> int:
        return self._num_rows - len(self._report.adcs_ports_in_use())

    @property
    def _num_analog_rows(self) -> int:
        return len(self._report.analog_output_ports_in_use()) * (1 + (self._samples is not None))

    @property
    def _num_digital_rows(self) -> int:
        return len(self._report.digital_output_ports_in_use()) * (1 + (self._samples is not None))

    def _is_row_analog(self, r) -> bool:
        return 1 <= r <= self._num_analog_rows

    def _is_row_digital(self, r) -> bool:
        return self._num_analog_rows < r <= self._num_output_rows

    def _is_row_analog_input(self, r) -> bool:
        return self._num_output_rows < r <= self._num_rows

    @property
    def _xrange(self) -> int:
        return (
            len(self._samples.analog["1"])
            if self._samples is not None
            else max(self._report.waveforms, key=lambda x: x.ends_at).ends_at + 100
        )

    def _pre_setup(self):
        self._setup_qe_colorscale()
        self._calculate_max_parallel_traces_per_row()
        return

    def _get_all_qe_used(self):
        return list(set([wf.element for wf in self._report.waveforms]))

    def _setup_qe_colorscale(self):
        qe_in_use = self._get_all_qe_used()
        n_colors = len(qe_in_use)
        samples = plotly.colors.qualitative.Pastel + plotly.colors.qualitative.Safe
        if n_colors > len(samples):
            samples += plotly.colors.sample_colorscale(
                "turbo",
                [n / (n_colors - len(samples)) for n in range(n_colors - len(samples))],
            )
        self._colormap = dict(zip(qe_in_use, samples))
        return

    def _calculate_max_parallel_traces_per_row(self):
        report_by_output_port = self._report.get_report_by_output_ports()
        self._max_parallel_traces_per_row["analog"] = {}
        self._max_parallel_traces_per_row["digital"] = {}

        def calc_row(waveform_list):
            max_in_row = 0
            functional_ts = sorted(
                [(r.timestamp, 1) for r in waveform_list] + [(r.ends_at, -1) for r in waveform_list],
                key=lambda t: t[0],
            )
            for _, f in functional_ts:
                max_in_row = max(max_in_row, max_in_row + f)
            return max_in_row

        for output_port, waveform_list in report_by_output_port["analog_out"].items():
            self._max_parallel_traces_per_row["analog"][output_port] = calc_row(waveform_list)

        for output_port, waveform_list in report_by_output_port["digital_out"].items():
            self._max_parallel_traces_per_row["digital"][output_port] = calc_row(waveform_list)

        return

    def _is_intersect(self, r1: Tuple[int, int], r2: Tuple[int, int]) -> bool:
        return (r1[0] <= r2[0] <= r1[1]) or (r1[0] <= r2[1] <= r1[1]) or (r2[0] < r1[0] and r2[1] > r1[1])

    def _get_hover_text(self, played_waveform):
        waveform_desc = played_waveform.to_string()
        if isinstance(played_waveform, PlayedAnalogWaveform):
            if played_waveform.chirp_info is not None:
                waveform_desc = played_waveform._to_custom_string(False)
                s = (
                    f"rate={played_waveform.chirp_info['rate']},units={played_waveform.chirp_info['units']},"
                    f" times={played_waveform.chirp_info['times']}\n"
                    + f"start_freq={pretty_string_freq(played_waveform.chirp_info['startFrequency'])}, "
                    + f"end_freq={pretty_string_freq(played_waveform.chirp_info['endFrequency'])}"
                )

                waveform_desc = f"<b>Chirp Pulse</b>\n({s})\n" + waveform_desc
        return "%{x}ns<br>" + waveform_desc.replace("\n", "</br>") + "<extra></extra>"

    def _get_output_port_waveform_plot_data(self, port_played_waveforms: List[PlayedWaveform], gui_args):
        graph_data = []
        annotations = []
        levels = []
        x_axis_name = gui_args["x_axis_name"]
        max_in_row = gui_args["max_in_row"]
        diff_between_traces, start_y = (0.2, 1.2) if max_in_row <= 7 else (1.4 / max_in_row, 1.45)
        y_level = [start_y] * 3
        for wf in port_played_waveforms:
            x_axis_points = (wf.timestamp, wf.ends_at)
            num_intersections = len([l for l in levels if self._is_intersect(l, x_axis_points)])
            levels.append(x_axis_points)
            prev_y = start_y if num_intersections == 0 else y_level[0]
            y_level = [prev_y - diff_between_traces] * 3
            graph_data.append(
                go.Scatter(
                    x=[x_axis_points[0], sum(x_axis_points) // 2, x_axis_points[1]],
                    y=y_level,
                    mode="lines+markers+text",
                    text=[
                        "",
                        f"{wf.pulse_name.removeprefix('OriginPulseName=')}"
                        + (f"({wf.get_iq_association})" if wf.is_iq else ""),
                        "",
                    ],
                    hovertemplate=self._get_hover_text(wf),
                    textfont=dict(size=10),
                    xaxis=x_axis_name,
                    name=wf.element,
                    legendgroup=wf.element,
                    showlegend=not (wf.element in self._already_registered_qe),
                    marker=dict(
                        line=dict(width=2, color=self._colormap[wf.element]),
                        symbol=["line-ns", "line-ew", "line-ns"],
                    ),
                    line=dict(color=self._colormap[wf.element], width=5),
                )
            )
            self._already_registered_qe.add(wf.element)

        return graph_data, annotations

    def _add_plot_data_for_analog_output_port(
        self, figure_row_number, output_port, port_waveforms: List[PlayedWaveform]
    ):
        if len(port_waveforms) == 0:
            return
        use_samples = self._samples is not None

        xaxis_name = f"x{figure_row_number}"
        port_wf_plot, annotations = self._get_output_port_waveform_plot_data(
            port_waveforms,
            {
                "x_axis_name": xaxis_name,
                "max_in_row": self._max_parallel_traces_per_row["analog"][output_port],
            },
        )
        row_number = figure_row_number * (1 + use_samples)
        self._figure.add_traces(port_wf_plot, rows=row_number, cols=1)
        [self._figure.add_annotation(annot, row=row_number, col=1) for annot in annotations]

        if use_samples:
            port_samples = self._samples.analog[str(output_port)]

            self._figure.add_trace(
                go.Scatter(
                    y=port_samples,
                    showlegend=False,
                    xaxis=xaxis_name,
                    hovertemplate="%{x}ns, %{y}v<extra></extra>",
                ),
                row=(figure_row_number * 2 - 1),
                col=1,
            )

        return

    def _add_plot_data_for_digital_output_port(
        self, figure_row_number, output_port, port_waveforms: List[PlayedWaveform]
    ):
        if len(port_waveforms) == 0:
            return
        use_samples = self._samples is not None

        xaxis_name = f"x{figure_row_number}"
        port_wf_plot, annotations = self._get_output_port_waveform_plot_data(
            port_waveforms,
            {
                "x_axis_name": xaxis_name,
                "max_in_row": self._max_parallel_traces_per_row["digital"][output_port],
            },
        )
        row_number = figure_row_number * (1 + use_samples)
        self._figure.add_traces(port_wf_plot, rows=row_number, cols=1)
        [self._figure.add_annotation(annot, row=row_number, col=1) for annot in annotations]

        if use_samples:
            port_samples = self._samples.digital.get(str(output_port), np.zeros(shape=self._xrange))
            if port_samples is None:
                logging.log(
                    logging.WARNING,
                    f"Could not find digital samples for output port {output_port}",
                )
            else:
                port_samples = port_samples.astype(int)
            samples_row_number = figure_row_number * 2 - 1
            self._figure.add_trace(
                go.Scatter(
                    y=port_samples,
                    showlegend=False,
                    xaxis=xaxis_name,
                    hovertemplate="%{x}ns, %{y}<extra></extra>",
                ),
                row=samples_row_number,
                col=1,
            )

        return

    def _add_plot_data_for_adc_port(
        self,
        figure_row_number,
        adc_port_number,
        adc_port_acquisitions: List[AdcAcquisition],
    ):
        graph_data = []
        levels = []
        y_level = [1.2] * 3
        for adc in adc_port_acquisitions:
            x_axis_points = (adc.start_time, adc.end_time)
            num_intersections = len([l for l in levels if self._is_intersect(l, x_axis_points)])
            levels.append(x_axis_points)
            prev_y = 1.2 if num_intersections == 0 else y_level[0]
            y_level = [prev_y - 0.2] * 3
            graph_data.append(
                go.Scatter(
                    x=[x_axis_points[0], sum(x_axis_points) // 2, x_axis_points[1]],
                    y=y_level,
                    mode="lines+markers+text",
                    text=[
                        "",
                        f"{adc.process}",
                        "",
                    ],
                    textfont=dict(size=10),
                    hovertemplate="%{x}ns<br>" + adc.to_string().replace("\n", "</br>") + "<extra></extra>",
                    name=adc.quantum_element,
                    legendgroup=adc.quantum_element,
                    showlegend=not (adc.quantum_element in self._already_registered_qe),
                    marker=dict(
                        line=dict(width=2, color=self._colormap[adc.quantum_element]),
                        symbol=["line-ns", "line-ew", "line-ns"],
                    ),
                    line=dict(color=self._colormap[adc.quantum_element], width=5),
                )
            )
            self._already_registered_qe.add(adc.quantum_element)

        self._figure.add_traces(graph_data, rows=figure_row_number, cols=1)

        return

    def _add_data(self):
        for (figure_row_number, (output_port, port_waveforms_list)) in enumerate(
            self._report.get_report_by_output_ports()["analog_out"].items()
        ):
            self._add_plot_data_for_analog_output_port(figure_row_number + 1, output_port, port_waveforms_list)

        for (figure_row_number, (output_port, port_waveforms_list)) in enumerate(
            self._report.get_report_by_output_ports()["digital_out"].items()
        ):
            self._add_plot_data_for_digital_output_port(
                figure_row_number + len(self._report.analog_output_ports_in_use()) + 1,
                output_port,
                port_waveforms_list,
            )

        for (figure_row_number, (input_port, adc_acquisition_list)) in enumerate(
            self._report.get_report_by_output_ports()["analog_in"].items()
        ):
            self._add_plot_data_for_adc_port(
                figure_row_number + self._num_output_rows + 1,
                input_port,
                adc_acquisition_list,
            )
        return

    def _update_extra_features(self):
        all_xaxis_names = sorted(
            [a for a in self._figure.layout.__dir__() if a.startswith("xaxis")],
            key=lambda s: int(s.removeprefix("xaxis")) if s.removeprefix("xaxis").isnumeric() else 0,
        )
        all_xaxis_names_short = {
            k: "x" + k.removeprefix("xaxis") if k.removeprefix("xaxis").isnumeric() else "" for k in all_xaxis_names
        }
        bottomost_x_axis = all_xaxis_names[-1]
        self._figure.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    direction="left",
                    active=0,
                    buttons=list(
                        [
                            dict(
                                args=[
                                    {k + ".matches": all_xaxis_names_short[bottomost_x_axis] for k in all_xaxis_names}
                                ],
                                label="Shared",
                                method="relayout",
                            ),
                            dict(
                                args=[{k + ".matches": v for k, v in all_xaxis_names_short.items()}],
                                label="Distinct",
                                method="relayout",
                            ),
                        ]
                    ),
                    showactive=True,
                    x=1,
                    xanchor="right",
                    y=1,
                    yanchor="bottom",
                    font=dict(size=10),
                ),
            ]
        )
        self._figure.add_annotation(
            dict(
                text="X-Axis scrolling method:",
                showarrow=False,
                x=1,
                y=1,
                yref="paper",
                yshift=40,
                yanchor="bottom",
                xref="paper",
                align="left",
            )
        )
        self._figure.update_layout(
            modebar_remove=[
                "autoscale",
                "autoscale2d",
                "lasso",
            ]
        )

        source_path = os.path.join(os.path.dirname(__file__), "sources", "logo_qm_square.png")

        im = base64.b64encode(open(source_path, "rb").read())
        self._figure.add_layout_image(
            source="data:image/png;base64,{}".format(im.decode()),
            xref="paper",
            yref="paper",
            x=0,
            y=1,
            sizex=0.1,
            sizey=0.1,
            xanchor="center",
            yanchor="bottom",
        )
        return

    def _setup_figure(self):
        self._figure = go.Figure()
        with_samples = self._samples is not None
        num_rows = max(self._num_rows, 4)
        titles = [f"Analog-Out-{a}" for a in self._report.analog_output_ports_in_use()] + [
            f"Digital-Out-{d}" for d in self._report.digital_output_ports_in_use()
        ]
        if with_samples:
            zipped = list(zip(titles, [[]] * self._num_output_rows))
            titles = [item for z in zipped for item in z]
        titles += [f"Analog-In-{a}" for a in self._report.adcs_ports_in_use()]

        if with_samples:
            specs = ([[{"t": 1.2 / (num_rows * 5)}]] + [[{"b": 1.2 / (num_rows * 5)}]]) * (
                self._num_output_rows // 2
            ) + [[{"t": 1 / (num_rows * 4)}]] * len(self._report.adcs_ports_in_use())
        else:
            specs = [[{"t": 1 / (num_rows * 4)}]] * num_rows
        self._figure.set_subplots(
            rows=num_rows,
            cols=1,
            subplot_titles=titles,
            vertical_spacing=0.1 / num_rows,
            specs=specs,
        )

        self._figure.update_layout(
            hovermode="closest",
            hoverdistance=5,
            height=160 * num_rows,
            title=dict(
                text=(
                    f"Waveform Report (connection: {self._report.controllers_in_use[0]})"
                    + (" for job: {}".format(self._job_id) if self._job_id != -1 else "")
                ),
                x=0.5,
                xanchor="center",
                yanchor="auto",
                xref="paper",
            ),
            legend=dict(title="Elements", y=0.98, yanchor="top"),
        )
        self._figure.add_annotation(
            dict(
                text=f"Created at {datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}",
                showarrow=False,
                x=0.5,
                y=1,
                yref="paper",
                yshift=20,
                yanchor="bottom",
                xref="paper",
                xanchor="center",
            )
        )

        for r in range(1, num_rows + 1):
            self._figure.update_xaxes(range=[0, self._xrange], row=r, col=1)

        if with_samples:
            for r in range(1, self._num_output_rows, 2):
                if self._is_row_digital(r):
                    sample_y_range = [-0.1, 1.1]
                else:
                    sample_y_range = [-0.6, 0.6]
                self._figure.update_yaxes(
                    range=sample_y_range,
                    row=r,
                    col=1,
                )
                self._figure.update_xaxes(showticklabels=False, row=r, col=1)
                title_text = "Voltage(v)"
                self._figure.update_yaxes(
                    title=dict(text=title_text, standoff=5, font=dict(size=9)),
                    row=r,
                    col=1,
                )

        for r in list(range(1 + with_samples, self._num_output_rows + 1, 1 + with_samples)) + [
            p + self._num_output_rows for p in self._report.adcs_ports_in_use()
        ]:
            self._figure.update_yaxes(
                range=[-0.5, 1.5],
                showticklabels=False,
                tickvals=[-0.5],
                ticklen=50,
                tickcolor="#000000",
                showgrid=False,
                zeroline=False,
                row=r,
                col=1,
            )
            self._figure.update_xaxes(title=dict(text="Time(ns)", standoff=5, font=dict(size=9)), row=r, col=1)
            title_text = ""

        return

    def build(self):
        self._pre_setup()
        self._setup_figure()
        self._add_data()
        self._update_extra_features()
        return

    def plot(self):
        if self._figure is None:
            raise RuntimeError("No graph has been built. Use 'build' on the instance to first build the figure.")
        self._figure.show()
        return

    def save(self, basedir="", filename=""):
        if not os.path.exists(basedir):
            os.makedirs(basedir)
        if filename == "":
            filename = f"waveform_report_{self._job_id}"
        if not os.path.splitext(filename)[1] == "html":
            filename += ".html"

        path = os.path.join(basedir, filename)
        with open(path, "w") as f:
            self._figure.write_html(f)
        return
